"""Test functions for the sparse.linalg._interface module
"""

import contextlib
from functools import partial
from itertools import product
import operator
from types import ModuleType
from typing import NamedTuple

import pytest
from pytest import raises as assert_raises, warns
from numpy.testing import assert_, assert_equal

import numpy as np
from scipy._lib._array_api import (
    SCIPY_ARRAY_API, SCIPY_DEVICE, is_torch, xp_assert_close, is_lazy_array,
    xp_assert_equal, xp_ravel, is_numpy, make_xp_pytest_marks, is_jax_array,
)
from scipy._external import array_api_extra as xpx
import scipy.sparse as sparse

import scipy.sparse.linalg._interface as interface
from scipy.sparse.linalg import LinearOperator, aslinearoperator
from scipy.sparse._sputils import matrix
from scipy._lib._gcutils import assert_deallocated

pytestmark = make_xp_pytest_marks(
    (LinearOperator, "__init__"),
    (LinearOperator, "matvec"),
    (LinearOperator, "matmat"),
    (LinearOperator, "dot"),
    (LinearOperator, "rdot"),
    (LinearOperator, "rmatvec"),
    (LinearOperator, "rmatmat"),
    aslinearoperator
)

def generate_broadcastable_shapes(nshapes, *, ndim=2, min=0, max=10, rng=None):
    rng = np.random.default_rng(rng)
    min = np.broadcast_to(min, ndim)  # so min and max can be scalars or array-like 
    max = np.broadcast_to(max, ndim)
    batch_shape = tuple(rng.integers(min_, max_+1) for min_, max_ in zip(min, max))
    shapes = np.repeat([batch_shape], nshapes, axis=0)

    # make some elements of some shapes 1 (while preserving overall batch shape)
    for column in shapes.T:
        column[rng.integers(1, nshapes):] = 1
    # permute elements between shapes (while preserving overall batch shape)
    shapes = list(rng.permuted(shapes, axis=0))
    # potentially trim preceeding 1s from a shape
    for i in range(len(shapes)):
        shape = shapes[i]
        j = np.where(shape != 1)[0][0] if np.any(shape != 1) else ndim
        if rng.random() < 0.25:
            shapes[i] = shape[rng.integers(j+1):]
            break

    assert np.broadcast_shapes(*shapes) == batch_shape
    return [tuple(int(el) for el in shape) for shape in shapes]


class TestLinearOperator:
    def setup_method(self):
        self.A = np.array([[1,2,3],
                           [4,5,6]])
        self.B = np.array([[1,2],
                           [3,4],
                           [5,6]])
        self.C = np.array([[1,2],
                           [3,4]])

    def test_matvec(self, xp):
        def get_matvecs(A):
            return [{
                'shape': A.shape,
                'dtype': xp.float64,
                'matvec': lambda x: A @ x,
                'rmatvec': lambda x: xp.conj(A.T) @ x,
                'rmatmat': lambda x: xp.conj(A.T) @ x,
                'matmat': lambda x: A @ x
            }]
        
        _asarray = partial(xp.asarray, dtype=xp.complex128)

        for matvecs in get_matvecs(_asarray(self.A)):
            A = interface.LinearOperator(**matvecs, xp=xp)

            assert A.args == ()

            xp_assert_equal(A.matvec(_asarray([1,2,3])), _asarray([14,32]))
            with pytest.warns(FutureWarning, match="column vectors"):
                xp_assert_equal(
                    A.matvec(_asarray([[1],[2],[3]])),
                    _asarray([[14],[32]]),
                )
            xp_assert_equal(A @ _asarray([1,2,3]), _asarray([14,32]))
            xp_assert_equal(A @ _asarray([[1],[2],[3]]), _asarray([[14],[32]]))
            xp_assert_equal(A.dot(_asarray([1,2,3])), _asarray([14,32]))
            xp_assert_equal(A.dot(_asarray([[1],[2],[3]])), _asarray([[14],[32]]))

            if not SCIPY_ARRAY_API:
                xp_assert_equal(
                    A.matvec(matrix([[1],[2],[3]])),
                    _asarray([[14],[32]]),
                )
                xp_assert_equal(A @ matrix([[1],[2],[3]]), _asarray([[14],[32]]))
                xp_assert_equal(A.dot(matrix([[1],[2],[3]])), _asarray([[14],[32]]))

            xp_assert_equal((2*A) @ _asarray([1,1,1]), _asarray([12,30]))
            xp_assert_equal((2 * A).rmatvec(_asarray([1, 1])), _asarray([10, 14, 18]))
            xp_assert_equal((2*A).H.matvec(_asarray([1,1])), _asarray([10, 14, 18]))
            xp_assert_equal(
                (2*A).adjoint().matvec(_asarray([1,1])),
                _asarray([10, 14, 18]),
            )
            xp_assert_equal((2*A) @ _asarray([[1],[1],[1]]), _asarray([[12],[30]]))
            xp_assert_equal(
                (2 * A).matmat(_asarray([[1], [1], [1]])),
                _asarray([[12], [30]]),
            )
            xp_assert_equal((A*2) @ _asarray([1,1,1]), _asarray([12,30]))
            xp_assert_equal((A*2) @ _asarray([[1],[1],[1]]), _asarray([[12],[30]]))
            xp_assert_equal((2j*A) @ _asarray([1,1,1]), _asarray([12j,30j]))
            xp_assert_equal((A+A) @ _asarray([1,1,1]), _asarray([12, 30]))
            xp_assert_equal((A + A).rmatvec(_asarray([1, 1])), _asarray([10, 14, 18]))
            xp_assert_equal((A+A).H.matvec(_asarray([1,1])), _asarray([10, 14, 18]))
            xp_assert_equal(
                (A + A).adjoint().matvec(_asarray([1,1])),
                _asarray([10, 14, 18]),
            )
            xp_assert_equal((A+A) @ _asarray([[1],[1],[1]]), _asarray([[12], [30]]))
            xp_assert_equal(
                (A+A).matmat(_asarray([[1],[1],[1]])),
                _asarray([[12], [30]])
            )
            xp_assert_equal((-A) @ _asarray([1,1,1]), _asarray([-6,-15]))
            xp_assert_equal((-A) @ _asarray([[1],[1],[1]]), _asarray([[-6],[-15]]))
            xp_assert_equal((A-A) @ _asarray([1,1,1]), _asarray([0,0]))
            xp_assert_equal((A - A) @ _asarray([[1], [1], [1]]), _asarray([[0], [0]]))

            A_ = xp.asarray(self.A, dtype=xp.complex128)
            X = xp.asarray([[1, 2], [3, 4]], dtype=xp.complex128)
            xp_assert_equal((2 * A).rmatmat(X), (2 * A_).T @ X)
            xp_assert_equal((A * 2).rmatmat(X), (A_ * 2).T @ X)
            xp_assert_equal((2j * A).rmatmat(X), xp.conj((2j * A_).T) @ X)
            xp_assert_equal((A * 2j).rmatmat(X), xp.conj((A_ * 2j).T) @ X)
            xp_assert_equal((A + A).rmatmat(X), (A_ + A_).T @ X)
            xp_assert_equal((A + 2j * A).rmatmat(X), xp.conj((A_ + 2j * A_).T) @ X)
            xp_assert_equal((-A).rmatmat(X), (-A_).T @ X)
            xp_assert_equal((A - A).rmatmat(X), (A_ - A_).T @ X)
            xp_assert_equal((2j * A).rmatmat(2j * X), xp.conj((2j * A_).T) @ (2j * X))

            z = A+A
            assert len(z.args) == 2 and z.args[0] is A and z.args[1] is A
            z = 2*A
            assert len(z.args) == 2
            if not is_lazy_array(xp.empty(0)):
                # lazy_xp_function breaks object identity
                assert z.args[0] is A and z.args[1] == 2

            array_object = type(xp.empty(0))
            assert isinstance(A.matvec(_asarray([1, 2, 3])), array_object)
            ctx = (
                pytest.warns(FutureWarning, match="column vectors")
                if not is_jax_array(xp.empty(0))
                else contextlib.nullcontext()
            )
            with ctx:
                assert isinstance(A.matvec(_asarray([[1],[2],[3]])), array_object)
            assert isinstance(A @ _asarray([1,2,3]), array_object)
            assert isinstance(A @ _asarray([[1],[2],[3]]), array_object)
            assert isinstance(A.dot(_asarray([1,2,3])), array_object)
            assert isinstance(A.dot(_asarray([[1],[2],[3]])), array_object)

            if not SCIPY_ARRAY_API:
                assert isinstance(A.matvec(matrix([[1],[2],[3]])), array_object)
                assert isinstance(A @ matrix([[1],[2],[3]]), array_object)
                assert isinstance(A.dot(matrix([[1],[2],[3]])), array_object)

            assert isinstance(2*A, interface._ScaledLinearOperator)
            assert isinstance(2j*A, interface._ScaledLinearOperator)
            assert isinstance(A+A, interface._SumLinearOperator)
            assert isinstance(-A, interface._ScaledLinearOperator)
            assert isinstance(A-A, interface._SumLinearOperator)
            assert isinstance(A/2, interface._ScaledLinearOperator)
            assert isinstance(A/2j, interface._ScaledLinearOperator)

            if not is_lazy_array(xp.empty(0)):
                # lazy_xp_function breaks object identity
                assert ((A * 3) / 3).args[0] is A  # check for simplification

            # Test that prefactor is of _ScaledLinearOperator is not mutated
            # when the operator is multiplied by a number
            result = A @ _asarray([1, 2, 3])
            B = A * 3
            C = A / 5
            xp_assert_equal(A @ _asarray([1, 2, 3]), result)

            assert (2j * A).dtype == xp.complex128

            # Test division by non-scalar
            msg = "Can only divide a linear operator by a scalar."
            with assert_raises(ValueError, match=msg):
                _ = A / xp.asarray([1, 2])

            assert_raises(ValueError, A.matvec, xp.asarray([1,2]))
            assert_raises(ValueError, A.matvec, xp.asarray([1,2,3,4]))
            assert_raises(ValueError, A.matvec, xp.asarray([[1],[2]]))
            assert_raises(ValueError, A.matvec, xp.asarray([[1],[2],[3],[4]]))

            assert_raises(ValueError, lambda: A@A)
            assert_raises(ValueError, lambda: A**2)

        for matvecsA, matvecsB in product(get_matvecs(_asarray(self.A)),
                                          get_matvecs(_asarray(self.B))):
            A = interface.LinearOperator(**matvecsA, xp=xp)
            B = interface.LinearOperator(**matvecsB, xp=xp)
            AtimesB = _asarray(self.A @ self.B)
            X = _asarray([[1, 2], [3, 4]])

            xp_assert_equal((A @ B).rmatmat(X), AtimesB.T @ X)
            xp_assert_equal((2j * A @ B).rmatmat(X), xp.conj((2j * AtimesB).T) @ X)

            xp_assert_equal((A @ B) @ _asarray([1,1]), _asarray([50,113]))
            xp_assert_equal((A @ B) @ _asarray([[1],[1]]), _asarray([[50],[113]]))
            xp_assert_equal((A @ B).matmat(_asarray([[1],[1]])), _asarray([[50],[113]]))

            xp_assert_equal((A @ B).rmatvec(_asarray([1, 1])), _asarray([71, 92]))
            xp_assert_equal((A @ B).H.matvec(_asarray([1, 1])), _asarray([71, 92]))
            xp_assert_equal(
                (A @ B).adjoint().matvec(_asarray([1, 1])),
                _asarray([71, 92]),
            )

            assert isinstance(A@B, interface._ProductLinearOperator)

            assert_raises(ValueError, lambda: A+B)
            assert_raises(ValueError, lambda: A**2)

            z = A@B
            assert len(z.args) == 2
            if not is_lazy_array(xp.empty(0)):
                # lazy_xp_function breaks object identity
                assert z.args[0] is A and z.args[1] is B

        for matvecsC in get_matvecs(_asarray(self.C)):
            C = interface.LinearOperator(**matvecsC, xp=xp)
            X = _asarray([[1, 2], [3, 4]])
            C_ = _asarray(self.C)

            xp_assert_equal(C.rmatmat(X), (C_).T @ X)
            xp_assert_equal((C**2).rmatmat(X), (C_ @ C_).T @ X)

            xp_assert_equal((C**2)@_asarray([1,1]), _asarray([17,37]))
            xp_assert_equal((C**2).rmatvec(_asarray([1, 1])), _asarray([22, 32]))
            xp_assert_equal((C**2).H.matvec(_asarray([1, 1])), _asarray([22, 32]))
            xp_assert_equal(
                (C**2).adjoint().matvec(_asarray([1, 1])),
                _asarray([22, 32]),
            )
            xp_assert_equal((C**2).matmat(_asarray([[1],[1]])), _asarray([[17],[37]]))

            assert isinstance(C**2, interface._PowerLinearOperator)

    @pytest.mark.skip_xp_backends(
        "array_api_strict",
        reason="https://github.com/data-apis/array-api-strict/issues/188",
    )
    def test_matmul(self, xp):
        _asarray = partial(xp.asarray, dtype=xp.complex128)
        A_ = _asarray(self.A)
        D = {'shape': A_.shape,
             'dtype': xp.complex128,
             'matvec': lambda x: xp.reshape(A_ @ x, (A_.shape[0],)),
             'rmatvec': lambda x: xp.reshape(xp.conj(A_.T) @ x, (A_.shape[1],)),
             'rmatmat': lambda x: xp.conj(A_.T) @ x,
             'matmat': lambda x: A_ @ x}
        A = interface.LinearOperator(**D, xp=xp)
        B = _asarray([[1 + 1j, 2, 3],
                      [4, 5, 6],
                      [7, 8, 9]])
        b = B[0, ...]

        xp_assert_equal(operator.matmul(A, b), A * b)
        xp_assert_equal(
            operator.matmul(A, xp.reshape(b, (-1, 1))),
            A * xp.reshape(b, (-1, 1)),
        )
        xp_assert_equal(operator.matmul(A, B), A @ B)
        xp_assert_equal(operator.matmul(b, A.H), b * A.H)
        xp_assert_equal(operator.matmul(b, A.adjoint()), b * A.adjoint())
        xp_assert_equal(
            operator.matmul(xp.reshape(b, (1, -1)), A.H),
            xp.reshape(b, (1, -1)) * A.H,
        )
        xp_assert_equal(
            operator.matmul(b.reshape(1, -1), A.adjoint()),
            b.reshape(1, -1) * A.adjoint(),
        )
        xp_assert_equal(operator.matmul(B, A.H), B @ A.H)
        xp_assert_equal(operator.matmul(B, A.adjoint()), B @ A.adjoint())
        assert_raises(ValueError, operator.matmul, A, 2)
        assert_raises(ValueError, operator.matmul, 2, A)

    def test_dimension_mismatch(self, xp):
        msg = "(D|d)imension mismatch"
        A = interface.aslinearoperator(xp.ones((3, 4)))
        # matvec
        A.matvec(xp.ones(4))
        with pytest.raises(ValueError, match=msg):
            A.matvec(xp.ones(3))
        # rmatvec
        A.rmatvec(xp.ones(3))
        with pytest.raises(ValueError, match=msg):
            A.rmatvec(xp.ones(4))
        # matmat
        A.matmat(xp.ones((4, 3)))
        with pytest.raises(ValueError, match=msg):
            A.matmat(xp.ones((3, 3)))
        # rmatmat
        A.rmatmat(xp.ones((3, 3)))
        with pytest.raises(ValueError, match=msg):
            A.rmatmat(xp.ones((4, 4)))
        # dot
        A.dot(xp.ones((4, 3)))
        with pytest.raises(ValueError, match=msg):
            A.dot(xp.ones((3, 3)))
        # rdot
        A.rdot(xp.ones((3, 3)))
        with pytest.raises(ValueError, match=msg):
            A.rdot(xp.ones((4, 4)))
        

@pytest.mark.skip_xp_backends("dask.array", reason="https://github.com/dask/dask/issues/11711")
class TestDotTests:
    """
    This class aims to help ensure correctness of the LinearOperator
    interface, by verifying correctness properties based on equivalent
    computations using 'forward' and 'adjoint' modes.
    """
    class OperatorArgs(NamedTuple):
        """
        shape: (core) shape of the operator
        op_dtype: dtype of the operator
        data_dtype: real dtype corresponding to op_dtype for data generation
        complex: the operator has a complex dtype
        """
        shape: tuple[int, ...]
        op_dtype: str
        data_dtype: str
        complex: bool

    real_square_args: OperatorArgs = OperatorArgs(
        (12, 12), "float64", "float64", False
    )
    integer_square_args: OperatorArgs = OperatorArgs(
        (9, 9), "int32", "float32", False
    )
    complex_square_args: OperatorArgs = OperatorArgs(
        (13, 13), "complex64", "float32", True
    )
    real_overdetermined_args: OperatorArgs = OperatorArgs(
        (17, 11), "float64", "float64", False
    )
    complex_overdetermined_args: OperatorArgs = OperatorArgs(
        (17, 11), "complex128", "float64", True
    )
    real_underdetermined_args: OperatorArgs = OperatorArgs(
        (5, 9), "float64", "float64", False
    )

    square_args_list: list[OperatorArgs] = [
        real_square_args, integer_square_args, complex_square_args
    ]
    all_args_list: list[OperatorArgs]  = square_args_list + [
        real_overdetermined_args, complex_overdetermined_args, real_underdetermined_args
    ]

    def check_matvec(
        self, xp: ModuleType, op: interface.LinearOperator, data_dtype: str,
        complex_data: bool = False,
    ):
        """
        This check verifies the equivalence of the forward and adjoint computation,
        using `matvec` and `rmatvec` respectively, on randomised data.

        Data is generated with the real dtype `data_dtype` and operated on by the
        linear operator `op`.

        If `complex_data` is set to `True`, complex data is instead generated
        by combining randomised real and imaginary components, each generated
        with `data_dtype`.

        If `check_operators` is set to `True`, equivalence is checked between
        `matvec` and `*` and `@`,
        and between `rmatvec` and the composition of `.H` and `*`.

        If `check_dot` is set to `True`, equivalence is checked between
        `matvec` and `.dot`,
        and between `rmatvec` and the composition of `.H` and `.dot`.
        """
        rng = np.random.default_rng(42)

        dtype = np.dtype(data_dtype)
        *batch_shape, M, N = op.shape

        u_broadcast, v_broadcast = generate_broadcastable_shapes(
            2, ndim=3, min=0, max=5, rng=63
        )

        for u_shape, v_shape in (
            # test broadcasting for unbatched RHS
            ((N,), (M,)),
            # as well as with the batch shape of `op` + 3-D broadcast dims
            ((*u_broadcast, *batch_shape, N), (*v_broadcast, *batch_shape, M))
        ):
            u = rng.standard_normal(u_shape, dtype=dtype)
            v = rng.standard_normal(v_shape, dtype=dtype)
            if complex_data:
                u = u + (1j * rng.standard_normal(u_shape, dtype=dtype))
                v = v + (1j * rng.standard_normal(v_shape, dtype=dtype))
            u = xp.asarray(u)
            v = xp.asarray(v)
    
            op_u = op.matvec(u)
            opH_v = op.rmatvec(v)
    
            op_u_H_v = xp.vecdot(op_u, v, axis=-1)
            uH_opH_v = xp.vecdot(u, opH_v, axis=-1)
    
            rtol = 1e-12 if np.finfo(data_dtype).eps < 1e-8 else 1e-5
            atol = 2e-15 if np.finfo(data_dtype).eps < 1e-8 else 1e-5
            xp_assert_close(op_u_H_v, uH_opH_v, rtol=rtol, atol=atol)

    def check_matmat(
        self, xp, op: interface.LinearOperator, data_dtype: str,
        complex_data: bool = False, check_operators: bool = False,
        check_dot: bool = False,
    ):
        """
        This check verifies the equivalence of the forward and adjoint computation,
        using `matmat` and `rmatmat` respectively, on randomised data.

        Data is generated with the real dtype `data_dtype` and operated on by the
        linear operator `op`.

        If `complex_data` is set to `True`, complex data is instead generated
        by combining randomised real and imaginary components, each generated
        with `data_dtype`.

        If `check_operators` is set to `True`, equivalence is checked between
        `matmat` and `*` and `@`,
        and between `rmatvec` and the composition of `.H` and `@`.

        If `check_dot` is set to `True`, equivalence is checked between
        `matmat` and `.dot`,
        and between `rmatmat` and the composition of `.H` and `.dot`.
        """
        rng = np.random.default_rng(42)
        k = rng.integers(2, 100)

        dtype = np.dtype(data_dtype)
        *batch_shape, M, N = op.shape

        # Test `U` and `V` with the same batch shape as `op`
        U = rng.standard_normal(size=(*batch_shape, N, k), dtype=dtype)
        V = rng.standard_normal(size=(*batch_shape, M, k), dtype=dtype)
        if complex_data:
            U = U + (1j * rng.standard_normal(size=(*batch_shape, N, k), dtype=dtype))
            V = V + (1j * rng.standard_normal(size=(*batch_shape, M, k), dtype=dtype))
        U = xp.asarray(U)
        V = xp.asarray(V)

        op_U = op.matmat(U)
        opH_V = op.rmatmat(V)

        if check_operators:
            xp_assert_close(op_U, op * U)
            xp_assert_close(op_U, op @ U)
            xp_assert_close(opH_V, op.H * V)
            xp_assert_close(opH_V, op.H @ V)

        if check_dot:
            xp_assert_close(op_U, op.dot(U))
            xp_assert_close(opH_V, op.H.dot(V))

        op_U_H = xp.conj(op_U).mT
        UH = xp.conj(U).mT

        op_U_H_V = xp.matmul(op_U_H, V)
        UH_opH_V = xp.matmul(UH, opH_V)

        rtol = 1e-12 if np.finfo(data_dtype).eps < 1e-8 else 1e-5
        atol = 2e-15 if np.finfo(data_dtype).eps < 1e-8 else 1e-5
        xp_assert_close(op_U_H_V, UH_opH_V, rtol=rtol, atol=atol)

    @pytest.mark.parametrize("batch_shape", [(), (3,), (3, 4, 5,), (0,)])
    @pytest.mark.parametrize("args", square_args_list)
    def test_identity_square(
        self, args: OperatorArgs, batch_shape: tuple[int, ...], xp
    ):
        """
        Simple identity operator on square matrices.
        Tests batches of RHS via `args.batch_shape`.
        """
        def identity(x):
            shape = (*xpx.broadcast_shapes(batch_shape, x.shape[:-1]), x.shape[-1])
            return xp.broadcast_to(x, shape)

        shape = batch_shape + args.shape
        dtype = getattr(xp, args.op_dtype)
        op = interface.LinearOperator(  # type:ignore[call-arg]
            shape=shape, dtype=dtype,
            matvec=identity, rmatvec=identity, xp=xp,
        )

        self.check_matvec(xp, op, data_dtype=args.data_dtype, complex_data=args.complex)
        self.check_matmat(xp, op, data_dtype=args.data_dtype, complex_data=args.complex)

    @pytest.mark.parametrize("batch_shape", [(), (3,), (3, 4, 5,), (0,)])
    @pytest.mark.parametrize("args", all_args_list)
    def test_identity_nonsquare(
        self, args: OperatorArgs, batch_shape: tuple[int, ...], xp
    ):
        """
        Identity operator with zero-padding on non-square matrices.
        Tests batches of RHS via `args.batch_shape`.
        """
        M, N = args.shape

        def pad_core_dim(x, target_size):
            pad_widths = [(0, 0)] * x.ndim # no padding for batch dims
            pad_widths[-1] = (0, target_size - x.shape[-1]) # padding for core dim
            return xpx.pad(x, pad_widths, mode="constant")

        def mv(x):
            match np.sign(x.shape[-1] - M):
                case 0:  # square
                    pass
                case 1:  # crop x to size
                    x = x[..., :M]
                case -1:  # pad with zeros
                    x = pad_core_dim(x, target_size=M)

            shape = (*xpx.broadcast_shapes(batch_shape, x.shape[:-1]), x.shape[-1])
            return xp.broadcast_to(x, shape)

        def rmv(x):            
            match np.sign(N - x.shape[-1]):
                case 0:  # square
                    pass
                case 1:  # pad with zeros
                    x = pad_core_dim(x, target_size=N)
                case -1:  # crop x to size
                    x = x[..., :N]

            shape = (*xpx.broadcast_shapes(batch_shape, x.shape[:-1]), x.shape[-1])
            return xp.broadcast_to(x, shape)

        shape = batch_shape + args.shape
        dtype = getattr(xp, args.op_dtype)
        op = interface.LinearOperator(  # type:ignore[call-arg]
            shape=shape, dtype=dtype, matvec=mv, rmatvec=rmv, xp=xp
        )
        
        self.check_matvec(xp, op, data_dtype=args.data_dtype, complex_data=args.complex)
        self.check_matmat(xp, op, data_dtype=args.data_dtype, complex_data=args.complex)

    @pytest.mark.parametrize("batch_shape", [(), (3,), (3, 4, 5,), (0,)])
    @pytest.mark.parametrize("args", square_args_list)
    def test_scaling_square(self, args: OperatorArgs, batch_shape: tuple[int, ...], xp):
        """
        Simple (complex) scaling operator on square matrices.
        Tests batches of RHS via `args.batch_shape`.
        """
        def scale(x):
            scale = (3 + 2j) / abs(3 + 2j)
            shape = (*xpx.broadcast_shapes(batch_shape, x.shape[:-1]), x.shape[-1])
            return xp.broadcast_to(scale * x, shape)

        def r_scale(x):
            scale = (3 - 2j) / abs(3 - 2j)
            shape = (*xpx.broadcast_shapes(batch_shape, x.shape[:-1]), x.shape[-1])
            return xp.broadcast_to(scale * x, shape)

        shape = batch_shape + args.shape
        dtype = getattr(xp, args.op_dtype)
        op = interface.LinearOperator(  # type:ignore[call-arg]
            shape=shape, dtype=dtype, matvec=scale, rmatvec=r_scale, xp=xp
        )
        self.check_matvec(xp, op, data_dtype=args.data_dtype, complex_data=args.complex)
        self.check_matmat(xp, op, data_dtype=args.data_dtype, complex_data=args.complex)

    @pytest.mark.parametrize("batch_shape", [(), (3,), (3, 4, 5,), (0,)])
    def test_subclass_matmat(self, batch_shape: tuple[int, ...], xp):
        """
        Simple rotation operator defined by `matmat` and `adjoint`,
        subclassing `LinearOperator`.
        Tests batches of RHS via `batch_shape`.
        """
        class RotOp(interface.LinearOperator):

            def __init__(self, dtype, shape, theta):
                self._theta = theta
                super().__init__(dtype, shape, xp=xp)

            def _matmat(self, X):
                theta = self._theta
                R = xp.asarray([
                    [np.cos(theta), -np.sin(theta)],
                    [np.sin(theta),  np.cos(theta)]
                ])
                B = R @ X
                shape = xpx.broadcast_shapes(batch_shape, X.shape[:-2]) + X.shape[-2:]
                return xp.broadcast_to(B, shape)

            def _adjoint(self):
                negative_theta = -self._theta
                return RotOp(self.dtype, self.shape, negative_theta)

        theta = xp.pi / 2
        data_dtype = "float64"
        op = RotOp(shape=(*batch_shape, 2, 2), dtype=xp.float64, theta=theta)

        self.check_matvec(xp, op, data_dtype=data_dtype, complex_data=False)
        self.check_matmat(xp, op, data_dtype=data_dtype, complex_data=False)

    @pytest.mark.parametrize("batch_shape", [(), (3,), (3, 4, 5,), (0,)])
    @pytest.mark.skip_xp_backends("dask.array", reason="https://github.com/dask/dask/issues/11711")
    def test_aslinearop_dense(
        self, batch_shape: tuple[int, ...], xp
    ):
        """
        Test operators coming from `aslinearoperator`,
        *including batched LHS*.
        """
        rng = np.random.default_rng(42)
        matrix = xp.asarray(
            rng.standard_normal((*batch_shape, 4, 4)),
            dtype=xp.complex128,
        )
        op = interface.aslinearoperator(matrix)
        data_dtype = "float64"
        self.check_matvec(xp, op, data_dtype=data_dtype, complex_data=True)
        self.check_matmat(xp, op, data_dtype=data_dtype, complex_data=True)

    @pytest.mark.parametrize("batch_shape", [(), (3,), (3, 4, 5,), (0,)])
    @pytest.mark.skip_xp_backends(np_only=True)
    def test_aslinearop_sparse(self, batch_shape: tuple[int, ...], xp):
        """
        Test operators coming from `aslinearoperator` with sparse arrays,
        *including batched LHS*
        """
        matrix = sparse.random_array((*batch_shape, 4, 4), dtype=xp.complex128)
        op = interface.aslinearoperator(matrix)
        data_dtype = "float64"
        self.check_matvec(xp, op, data_dtype=data_dtype, complex_data=True)
        self.check_matmat(xp, op, data_dtype=data_dtype, complex_data=True)


class TestAsLinearOperator:
    def make_cases(self, original, dtype, xp=None):
        cases = []

        # Test default implementations of _adjoint and _rmatvec, which
        # refer to each other.
        _xp = xp if xp is not None else np
        def mv(x):
            y = original @ x
            if len(x.shape) == 2:
                y = _xp.reshape(y, (-1, 1))
            return y

        def rmv(x):
            return _xp.conj(original.T) @ x

        class BaseMatlike(interface.LinearOperator):
            args = ()

            def __init__(self, dtype):
                super().__init__(
                    dtype=dtype,
                    shape=original.shape,
                    xp=xp,
                )

            def _matvec(self, x):
                return mv(x)

        class HasRmatvec(BaseMatlike):
            args = ()

            def _rmatvec(self,x):
                return rmv(x)

        class HasAdjoint(BaseMatlike):
            args = ()

            def _adjoint(self):
                shape = self.shape[1], self.shape[0]
                matvec = partial(rmv)
                rmatvec = partial(mv)
                return interface.LinearOperator(matvec=matvec,
                                                rmatvec=rmatvec,
                                                dtype=dtype,
                                                shape=shape,
                                                xp=xp,)

        class HasRmatmat(HasRmatvec):
            def _matmat(self, x):
                return original @ x

            def _rmatmat(self, x):
                return _xp.conj(original.T) @ x

        if xp:
            cases.append((original, original))

            cases.append((HasRmatvec(dtype), original))
            cases.append((HasAdjoint(dtype), original))
            cases.append((HasRmatmat(dtype), original))
            
            cases.append((interface.aslinearoperator(original.T).T, original))
            cases.append((
                interface.aslinearoperator(original.T).H,
                xp.conj(original),
            ))
            cases.append((
                interface.aslinearoperator(original.T).adjoint(),
                xp.conj(original)),
            )

            return cases
        
        cases.append((matrix(original, dtype=dtype), original))
        cases.append((np.array(original, dtype=dtype), original))
        cases.append((sparse.csr_array(original, dtype=dtype), original))

        cases.append((HasRmatvec(dtype), original))
        cases.append((HasAdjoint(dtype), original))
        cases.append((HasRmatmat(dtype), original))
        return cases
    
    def setup_method(self):
        self.cases = []
        make_cases = self.make_cases
        original = np.array([[1,2,3], [4,5,6]])
        self.cases += make_cases(original, np.int32)
        self.cases += make_cases(original, np.float32)
        self.cases += make_cases(original, np.float64)
        self.cases += [(interface.aslinearoperator(M).T, A.T)
                       for M, A in make_cases(original.T, np.float64)]
        self.cases += [(interface.aslinearoperator(M).H, A.T.conj())
                       for M, A in make_cases(original.T, np.float64)]
        self.cases += [(interface.aslinearoperator(M).adjoint(), A.T.conj())
                       for M, A in make_cases(original.T, np.float64)]

        original = np.array([[1, 2j, 3j], [4j, 5j, 6]])
        self.cases += make_cases(original, np.complex128)
        self.cases += [(interface.aslinearoperator(M).T, A.T)
                       for M, A in make_cases(original.T, np.complex128)]
        self.cases += [(interface.aslinearoperator(M).H, A.T.conj())
                       for M, A in make_cases(original.T, np.complex128)]
        self.cases += [(interface.aslinearoperator(M).adjoint(), A.T.conj())
                       for M, A in make_cases(original.T, np.complex128)]

    @pytest.mark.skip_xp_backends(np_only=True)
    def test_basic(self, xp):

        for M, A_array in self.cases:
            A = interface.aslinearoperator(M)

            xs = [np.array([1, 2, 3]),
                  np.array([[1], [2], [3]])]
            ys = [np.array([1, 2]), np.array([[1], [2]])]

            if A.dtype == np.complex128:
                xs += [np.array([1, 2j, 3j]),
                       np.array([[1], [2j], [3j]])]
                ys += [np.array([1, 2j]), np.array([[1], [2j]])]

            x2 = np.array([[1, 4], [2, 5], [3, 6]])

            for x in xs:
                ctx = (
                    pytest.warns(FutureWarning, match="column vectors")
                    if x.shape[-1] == 1
                    else contextlib.nullcontext()
                )
                with ctx:
                    assert_equal(A.matvec(x), A_array.dot(x))

                assert_equal(A @ x, A_array.dot(x))

            assert_equal(A.matmat(x2), A_array.dot(x2))
            assert_equal(A @ x2, A_array.dot(x2))

            for y in ys:
                ctx = (
                    pytest.warns(FutureWarning, match="column vectors")
                    if y.shape[-1] == 1
                    else contextlib.nullcontext()
                )
                with ctx:
                    assert_equal(A.rmatvec(y), A_array.T.conj().dot(y))
                    assert_equal(A.T.matvec(y), A_array.T.dot(y))
                    assert_equal(A.H.matvec(y), A_array.T.conj().dot(y))
                    assert_equal(A.adjoint().matvec(y), A_array.T.conj().dot(y))

            for y in ys:
                if y.ndim < 2:
                    continue
                assert_equal(A.rmatmat(y), A_array.T.conj().dot(y))
                assert_equal(A.T.matmat(y), A_array.T.dot(y))
                assert_equal(A.H.matmat(y), A_array.T.conj().dot(y))
                assert_equal(A.adjoint().matmat(y), A_array.T.conj().dot(y))

            if hasattr(M,'dtype'):
                assert_equal(A.dtype, M.dtype)

            assert_(hasattr(A, 'args'))

    @pytest.mark.skip_xp_backends(np_only=True)
    def test_dot(self, xp):

        for M, A_array in self.cases:
            A = interface.aslinearoperator(M)

            x0 = np.array([1, 2, 3])
            x1 = np.array([[1], [2], [3]])
            x2 = np.array([[1, 4], [2, 5], [3, 6]])

            assert_equal(A.dot(x0), A_array.dot(x0))
            assert_equal(A.dot(x1), A_array.dot(x1))
            assert_equal(A.dot(x2), A_array.dot(x2))

    @pytest.mark.parametrize("dtype", ["int64", "float64", "complex128"])
    def test_xp(self, dtype, xp):
        if dtype == "int64" and is_torch(xp) and SCIPY_DEVICE != "cpu":
            pytest.skip("\"addmm_cuda\" not implemented for 'Long'")
        dtype = getattr(xp, dtype)
        original = xp.asarray([[1, 2, 3], [4, 5, 6]], dtype=dtype)
        for M, A_array in self.make_cases(original, dtype, xp=xp):
            A = interface.aslinearoperator(M)

            xs = [xp.asarray([1, 2, 3]),
                  xp.asarray([[1], [2], [3]])]
            ys = [xp.asarray([1, 2]), xp.asarray([[1], [2]])]

            if A.dtype == xp.complex128:
                xs += [xp.asarray([1, 2j, 3j]),
                       xp.asarray([[1], [2j], [3j]])]
                ys += [xp.asarray([1, 2j]), xp.asarray([[1], [2j]])]

            x2 = xp.asarray([[1, 4], [2, 5], [3, 6]], dtype=dtype)

            for x in xs:
                x = xp.astype(x, dtype)
                xp_assert_equal(A @ x, A_array @ x)

            xp_assert_equal(A.matmat(x2), A_array @ x2)
            xp_assert_equal(A @ x2, A_array @ x2)

            for y in ys:
                if y.ndim < 2:
                    continue
                y = xp.astype(y, dtype)
                xp_assert_equal(A.rmatmat(y), xp.conj(A_array.T) @ y)
                xp_assert_equal(A.T.matmat(y), xp.conj(A_array.T) @ y)
                xp_assert_equal(A.H.matmat(y), xp.conj(A_array.T) @ y)
                xp_assert_equal(A.adjoint().matmat(y), xp.conj(A_array.T) @ y)

            if hasattr(M, 'dtype'):
                assert_equal(A.dtype, M.dtype)

            assert hasattr(A, 'args')
            

def test_repr(xp):
    A = interface.LinearOperator(shape=(1, 1), matvec=lambda x: xp.asarray([1]), xp=xp)
    repr_A = repr(A)
    assert 'unspecified dtype' not in repr_A, repr_A


def test_identity(xp):
    ident = interface.IdentityOperator((3, 3), xp=xp)
    xp_assert_equal(ident @ xp.asarray([1, 2, 3]), xp.asarray([1, 2, 3]))
    xp_assert_equal(xp_ravel(ident.dot(xp.reshape(xp.arange(9), (3, 3)))), xp.arange(9))

    assert_raises(ValueError, ident.matvec, xp.asarray([1, 2, 3, 4]))


def test_attributes(xp):
    A = interface.aslinearoperator(xp.reshape(xp.arange(16, dtype=xp.float64), (4, 4)))

    def always_four_ones(x):
        x = xp.asarray(x)
        assert x.shape == (3,) or x.shape == (3, 1)
        return xp.ones(4)

    B = interface.LinearOperator(shape=(4, 3), matvec=always_four_ones, xp=xp)

    ops = [A, B, A * B, A @ B, A.H, A.adjoint(), A + A, B + B, A**4]
    for op in ops:
        assert hasattr(op, "dtype")
        assert hasattr(op, "shape")
        assert hasattr(op, "_matvec")


def matvec_for_pickle(x):
    """ Needed for test_pickle as local functions are not pickleable """
    return x


@pytest.mark.skip_xp_backends(
    "array_api_strict",
    reason="pickle-ability is not guaranteed by the standard"
)
def test_pickle(xp):
    import pickle

    protocol_min = 0 if is_numpy(xp) else 2
    for protocol in range(protocol_min, pickle.HIGHEST_PROTOCOL + 1):
        A = interface.LinearOperator((3, 3), matvec_for_pickle, xp=xp)
        s = pickle.dumps(A, protocol=protocol)
        B = pickle.loads(s)

        for k in A.__dict__:
            assert getattr(A, k) == getattr(B, k)


def test_inheritance(xp):
    class Empty(interface.LinearOperator):
        pass

    with warns(RuntimeWarning, match="should implement at least"):
        assert_raises(TypeError, Empty)

    class Identity(interface.LinearOperator):
        def __init__(self, n):
            super().__init__(dtype=None, shape=(n, n), xp=xp)

        def _matvec(self, x):
            return x

    id3 = Identity(3)
    xp_assert_equal(id3.matvec(xp.asarray([1, 2, 3])), xp.asarray([1, 2, 3]))
    assert_raises(NotImplementedError, id3.rmatvec, xp.asarray([4, 5, 6]))

    class MatmatOnly(interface.LinearOperator):
        def __init__(self, A):
            super().__init__(A.dtype, A.shape, xp=xp)
            self.A = A

        def _matmat(self, x):
            return self.A @ x

    mm = MatmatOnly(xp.asarray(np.random.randn(5, 3)))
    assert mm.matvec(xp.asarray(np.random.randn(3))).shape == (5,)


def test_dtypes_of_operator_sum(xp):
    # gh-6078

    mat_complex = xp.asarray(np.random.rand(2,2) + 1j * np.random.rand(2,2))
    mat_real = xp.asarray(np.random.rand(2,2))

    complex_operator = interface.aslinearoperator(mat_complex)
    real_operator = interface.aslinearoperator(mat_real)

    sum_complex = complex_operator + complex_operator
    sum_real = real_operator + real_operator

    assert sum_real.dtype == xp.float64
    assert sum_complex.dtype == xp.complex128


def test_no_double_init(xp):
    call_count = [0]

    def matvec(v):
        call_count[0] += 1
        return v

    # It should call matvec exactly once (in order to determine the
    # operator dtype)
    interface.LinearOperator((2, 2), matvec=matvec, xp=xp)
    assert_equal(call_count[0], 1)


INT_DTYPES = (np.int8, np.int16, np.int32, np.int64)
REAL_DTYPES = (np.float32, np.float64, np.longdouble)
COMPLEX_DTYPES = (np.complex64, np.complex128, np.clongdouble)
INEXACTDTYPES = REAL_DTYPES + COMPLEX_DTYPES
ALLDTYPES = INT_DTYPES + INEXACTDTYPES


@pytest.mark.parametrize("test_dtype", ALLDTYPES)
def test_determine_lo_dtype_from_matvec(test_dtype, xp):
    if "longdouble" in test_dtype.__name__ and not is_numpy(xp):
        pytest.skip("longdoubles are only tested for `np`")
    # gh-19209
    scalar = xp.asarray(np.array(1, dtype=test_dtype))
    def mv(v):
        return xp.stack([scalar * v[0], v[1]])

    lo = interface.LinearOperator((2, 2), matvec=mv, xp=xp)
    # expected dtype depends on if mixed exact-inexact promotion is defined
    # since dtype determination follows the following procedure:
    # - take the dtype from calling `matvec` on an `int8`
    # - unless that fails (e.g. via overflow or no mixed exact-inexact promotion),
    #   in which case use the default integral dtype
    expected = scalar.dtype
    if xp.isdtype(expected, ("real floating", "complex floating")):
        try:
            xp.asarray(2) + xp.asarray(2.0)
        except TypeError:
            expected = xpx.default_dtype(xp, kind="integral")
    assert lo.dtype == expected


def test_determine_lo_dtype_for_int(xp):
    # gh-19209
    # test Python int larger than int8 max cast to some int
    def mv(v):
        return xp.asarray([128 * v[0], v[1]])

    lo = interface.LinearOperator((2, 2), matvec=mv, xp=xp)
    assert xp.isdtype(lo.dtype, "integral")


def test_adjoint_conjugate(xp):
    X = xp.asarray([[1j]], dtype=xp.complex128)
    A = interface.aslinearoperator(X)

    B = 1j * A
    Y = 1j * X

    v = xp.asarray([1], dtype=xp.complex128)

    xp_assert_equal(B.dot(v), xp.vecdot(Y, v))
    xp_assert_equal(B.H.dot(v), xp.vecdot(xp.conj(Y.T), v))
    xp_assert_equal(B.adjoint().dot(v), xp.vecdot(xp.conj(Y.T), v))


def test_ndim(xp):
    X = xp.asarray([[1]])
    A = interface.aslinearoperator(X)
    assert A.ndim == 2


def test_transpose_noconjugate(xp):
    X = xp.asarray([[1j]], dtype=xp.complex128)
    A = interface.aslinearoperator(X)

    B = 1j * A
    Y = 1j * X

    v = xp.asarray([1], dtype=xp.complex128)

    xp_assert_equal(B.dot(v), xp.vecdot(Y, v))
    xp_assert_equal(B.T.dot(v), xp.vecdot(Y.T, v))


@pytest.mark.skip_xp_backends("dask.array", reason="https://github.com/dask/dask/issues/11711")
@pytest.mark.skip_xp_backends(
    "array_api_strict",
    reason="https://github.com/data-apis/array-api-strict/issues/188"
)
def test_transpose_multiplication(xp):
    _asarray = partial(xp.asarray, dtype=xp.complex128)
    class MyMatrix(interface.LinearOperator):
        def __init__(self, A):
            super().__init__(A.dtype, A.shape, xp=xp)
            self.A = A
        def _matmat(self, other): return self.A @ other
        def _rmatmat(self, other): return self.A.mT @ other

    A = MyMatrix(_asarray([[1, 2], [3, 4]]))
    X = _asarray([1, 2])
    X_T = X
    B = _asarray([[10, 20], [30, 40]])
    X2 = xp.reshape(X, (-1, 1))
    Y = _asarray([[1, 2], [3, 4]])

    xp_assert_equal(A @ B, Y @ B)
    xp_assert_equal(B.T @ A, B.T @ Y)
    xp_assert_equal(A.T @ B, Y.mT @ B)
    xp_assert_equal(A @ X, Y @ X)
    xp_assert_equal(X_T @ A, X_T @ Y)
    xp_assert_equal(A.T @ X, Y.mT @ X)
    xp_assert_equal(A @ X2, Y @ X2)
    xp_assert_equal(X2.mT @ A, X2.mT @ Y)
    xp_assert_equal(A.T @ X2, Y.mT @ X2)


@pytest.mark.skip_xp_backends(np_only=True)
def test_sparse_matmat_exception():
    B = sparse.eye_array(2)
    # well defined matmat via `aslinearoperator`
    A = interface.aslinearoperator(sparse.eye_array(2))
    assert isinstance(A @ B, sparse.sparray)
    assert isinstance(B @ A, sparse.sparray)
    xp_assert_equal((A @ B).toarray(), np.eye(2))
    xp_assert_equal((B @ A).toarray(), np.eye(2))
    # ill-defined matmat via default fallback to matvec
    A = interface.LinearOperator((2, 2), matvec=lambda x: x)
    msg = "Try wrapping the matrix with `aslinearoperator` first."
    with assert_raises(TypeError, match=msg):
        A @ B
    with assert_raises(TypeError, match=msg):
        B @ A
    # after using `aslinearoperator`
    B = interface.aslinearoperator(B)
    assert isinstance(A @ B, interface.LinearOperator)
    assert isinstance(B @ A, interface.LinearOperator)
    xp_assert_equal((A @ B).matvec(np.ones(2)), np.ones(2))


@pytest.mark.skip_xp_backends(
    "jax.numpy", reason="remaining reference (due to lazy_xp_function?)"
)
def test_MatrixLinearOperator_refcycle(xp):
    # gh-10634
    # Test that MatrixLinearOperator can be automatically garbage collected
    A = xp.eye(2)
    with assert_deallocated(interface.MatrixLinearOperator, A, xp) as op:
        op.adjoint()
        del op


@pytest.mark.parametrize("left", [False, True])
@pytest.mark.parametrize("operator_definition", ["subclass_matvec", "subclass_matmat",
                                                 "__init__matvec", "__init__matmat",
                                                 "aslinearoperator"])
@pytest.mark.parametrize("batch_A", [(), (5,), (0,)])
@pytest.mark.parametrize("batch_x", [(), (6, 1), (0, 1)])
@pytest.mark.parametrize("dtype", [np.float64, np.complex128])
@pytest.mark.skip_xp_backends("dask.array", reason="https://github.com/dask/dask/issues/11711")
@pytest.mark.skip_xp_backends(
    "array_api_strict",
    reason="https://github.com/data-apis/array-api-strict/issues/188"
)
def test_batch(left, operator_definition, batch_A, batch_x, dtype, xp):
    # TODO ideas:
    # - test lower-precision types
    # - test `transpose`, `adjoint`, `__mul__`, etc.
    # - test composite LinearOperators
    rng = np.random.default_rng(41981392342349823)

    m, n, k = 4, 3, 2
    A_ = rng.random((*batch_A, m, n))
    x_row = rng.random((*batch_x, n if left else m))
    x_col = rng.random((*batch_x, n if left else m, 1))
    x_mat = rng.random((*batch_x, n if left else m, k))

    if dtype == np.complex128:
        A_ = A_ + 1j * rng.random(A_.shape)
        x_row = x_row + 1j * rng.random(x_row.shape)
        x_col = x_col + 1j * rng.random(x_col.shape)
        x_mat = x_mat + 1j * rng.random(x_mat.shape)
    
    A_, x_row, x_col, x_mat = (xp.asarray(x) for x in (A_, x_row, x_col, x_mat))

    def matvec(A, x):
        if not batch_x:
            assert x.ndim == 1
            return A @ x
        else:
            # It might make it easier on the author of LinearOperators to
            # "ravel" all batch dimensions into one before calling their `matvec`
            # implementation. That way, they only have to think about vectorizing
            # w.r.t. one batch dimension rather than an arbitrary number of batch
            # dimensions. In this case, `x.ndim` would be exactly 2.
            assert x.ndim >= 2
            return (A @ (x[..., xp.newaxis]))[..., 0]

    def matmat(A, X):
        if not batch_x:
            assert X.ndim == 2
        else:
            # Similar to above, we could ravel all batch dimensions before calling
            # the user's `matmat`. Then `x.ndim` would be exactly 3.
            assert X.ndim >= 3
        return A @ X

    if operator_definition == "aslinearoperator":
        A = interface.aslinearoperator(A_)
    elif operator_definition == "__init__matvec":
        A = interface.LinearOperator(
            shape=A_.shape, dtype=A_.dtype,
            matvec=lambda x: matvec(A_, x),
            rmatvec=lambda x: matvec(xp.conj(A_.mT), x),
            xp=xp,
        )
    elif operator_definition == "__init__matmat":
        # A = interface.LinearOperator(shape=A_.shape, dtype=A_.dtype,
        #                              matmat=lambda X: matmat(A_, X))
        pytest.skip("should work but doesn't - see gh-24510")
    elif operator_definition == "subclass_matvec":
        class MyLinearOperator(interface.LinearOperator):
            def _matvec(self, x):
                return matvec(A_, x)
            def _rmatvec(self, x):
                return matvec(xp.conj(A_.mT), x)
        A = MyLinearOperator(shape=A_.shape, dtype=A_.dtype, xp=xp)
    elif operator_definition == "subclass_matmat":
        class MyLinearOperator(interface.LinearOperator):
            def _matmat(self, X):
                return matmat(A_, X)
            def _rmatmat(self, X):
                return matmat(xp.conj(A_.mT), X)
        A = MyLinearOperator(shape=A_.shape, dtype=A_.dtype, xp=xp)

    # Test matvec
    # a. with row vector (or batch of row vectors)
    xp_assert_close(
        A.matvec(x_row) if left else A.rmatvec(x_row),
        matvec(A_, x_row) if left else matvec(xp.conj(A_.mT), x_row)
    )
    # b. with column vector (or batch of column vectors)
    if batch_x:
        message = "Dimension mismatch:..."
        with pytest.raises(ValueError, match=message):
            A.matvec(x_col) if left else A.rmatvec(x_col)
    else:
        message = ("Calling `matvec` on 'column vectors'..." if left
                   else "Calling `rmatvec` on 'column vectors'...")
        with pytest.warns(FutureWarning, match=message):
            xp_assert_close(
                A.matvec(x_col) if left else A.rmatvec(x_col),
                A_ @ x_col if left else xp.conj(A_.mT) @ x_col,
            )
    # c. with matrix (or batch of matrices)
    with pytest.raises(ValueError, match="Dimension mismatch:..."):
        A.matvec(x_mat) if left else A.rmatvec(x_mat)

    # Test matmat
    # a. with row vector (or batch of row vectors)
    message = "Dimension mismatch:..." if batch_x else "Expected at least 2-d..."
    with pytest.raises(ValueError, match=message):
        A.matmat(x_row) if left else A.rmatmat(x_row)
    # b. with column vector (or batch of column vectors)
    xp_assert_close(
        A.matmat(x_col) if left else A.rmatmat(x_col),
        A_ @ x_col if left else xp.conj(A_.mT) @ x_col,
    )
    # c. with matrix (or batch of matrices)
    xp_assert_close(
        A.matmat(x_mat) if left else A.rmatmat(x_mat),
        A_ @ x_mat if left else xp.conj(A_.mT) @ x_mat,
    )

    # test __matmul__ (via `@`), `__call__`, and `dot`
    if left:
        # a. with row vector (or batch of row vectors)
        if batch_x:
            with pytest.raises(ValueError, match="Dimension mismatch:..."):
                A @ x_row
        else:
            xp_assert_close(A @ x_row, A_ @ x_row)
            xp_assert_close(A(x_row), A_ @ x_row)
            xp_assert_close(A.dot(x_row), A_ @ x_row)
        # b. with column vector (or batch of column vectors)
        xp_assert_close(A @ x_col, A_ @ x_col)
        xp_assert_close(A(x_col), A_ @ x_col)
        xp_assert_close(A.dot(x_col), A_ @ x_col)
        # c. with matrix (or batch of matrices)
        xp_assert_close(A @ x_mat, A_ @ x_mat)
        xp_assert_close(A(x_mat), A_ @ x_mat)
        xp_assert_close(A.dot(x_mat), A_ @ x_mat)
    else:
        # a. with row vector (or batch of row vectors)
        broadcastable = True
        try:
            # `batch_x[-1]` becomes part of the core shape for matrix
            # multiplication with 1-D row vector
            # success thus depends on whether `batch_A` can broadcast
            # with `batch_x[:-1]`
            np.broadcast_shapes(batch_A, batch_x[:-1])
        except ValueError:
            broadcastable = False
        if not broadcastable:
            # Maybe should be "Dimension mismatch..."?
            msg = "Incompatible shapes|size of tensor|could not be broadcast..."
            with pytest.raises((ValueError, RuntimeError), match=msg):
                x_row @ A
        else:
            xp_assert_close(x_row @ A, x_row @ A_)
            xp_assert_close(A.rdot(x_row), x_row @ A_)
        # b. with column vector (or batch of column vectors)
        with pytest.raises(ValueError, match="Dimension mismatch:..."):
            x_col @ A
        # c. with matrix (or batch of matrices)
        xp_assert_close(xp.conj(x_mat.mT) @ A, xp.conj(x_mat.mT) @ A_)
        xp_assert_close(A.rdot(xp.conj(x_mat.mT)), xp.conj(x_mat.mT) @ A_)



"""Tools for testing implementations of __array_function__ and ufunc overrides


"""

import numpy._core.umath as _umath
from numpy import ufunc as _ufunc
from numpy._core.overrides import ARRAY_FUNCTIONS as _array_functions


def get_overridable_numpy_ufuncs():
    """List all numpy ufuncs overridable via `__array_ufunc__`

    Parameters
    ----------
    None

    Returns
    -------
    set
        A set containing all overridable ufuncs in the public numpy API.
    """
    ufuncs = {obj for obj in _umath.__dict__.values()
              if isinstance(obj, _ufunc)}
    return ufuncs


def allows_array_ufunc_override(func):
    """Determine if a function can be overridden via `__array_ufunc__`

    Parameters
    ----------
    func : callable
        Function that may be overridable via `__array_ufunc__`

    Returns
    -------
    bool
        `True` if `func` is overridable via `__array_ufunc__` and
        `False` otherwise.

    Notes
    -----
    This function is equivalent to ``isinstance(func, np.ufunc)`` and
    will work correctly for ufuncs defined outside of Numpy.

    """
    return isinstance(func, _ufunc)


def get_overridable_numpy_array_functions():
    """List all numpy functions overridable via `__array_function__`

    Parameters
    ----------
    None

    Returns
    -------
    set
        A set containing all functions in the public numpy API that are
        overridable via `__array_function__`.

    """
    # 'import numpy' doesn't import recfunctions, so make sure it's imported
    # so ufuncs defined there show up in the ufunc listing
    from numpy.lib import recfunctions  # noqa: F401
    return _array_functions.copy()

def allows_array_function_override(func):
    """Determine if a Numpy function can be overridden via `__array_function__`

    Parameters
    ----------
    func : callable
        Function that may be overridable via `__array_function__`

    Returns
    -------
    bool
        `True` if `func` is a function in the Numpy API that is
        overridable via `__array_function__` and `False` otherwise.
    """
    return func in _array_functions



from _typeshed import Incomplete
from collections.abc import Callable, Sequence
from typing import (
    Any,
    Concatenate,
    Never,
    Protocol,
    Self,
    SupportsIndex,
    overload,
    type_check_only,
)

import numpy as np
from numpy._typing import (
    ArrayLike,
    NDArray,
    _AnyShape,
    _ArrayLike,
    _ArrayLikeBool_co,
    _ArrayLikeComplex_co,
    _ArrayLikeFloat_co,
    _ArrayLikeInt,
    _ArrayLikeInt_co,
    _ArrayLikeObject_co,
    _ArrayLikeUInt_co,
    _Shape,
    _ShapeLike,
)

__all__ = [
    "column_stack",
    "dstack",
    "array_split",
    "split",
    "hsplit",
    "vsplit",
    "dsplit",
    "apply_over_axes",
    "expand_dims",
    "apply_along_axis",
    "kron",
    "tile",
    "take_along_axis",
    "put_along_axis",
]

# Signature of `__array_wrap__`
@type_check_only
class _ArrayWrap(Protocol):
    def __call__(
        self,
        array: NDArray[Any],
        context: tuple[np.ufunc, tuple[Any, ...], int] | None = ...,
        return_scalar: bool = ...,
        /,
    ) -> Any: ...

@type_check_only
class _SupportsArrayWrap(Protocol):
    @property
    def __array_wrap__(self) -> _ArrayWrap: ...

# Protocol for array-like objects that preserve their type through split operations.
# Requires shape for size, ndim for dimensional checks in hsplit/vsplit/dsplit,
# swapaxes for axis manipulation, and __getitem__ for slicing.
@type_check_only
class _SupportsSplitOps(Protocol):
    @property
    def shape(self) -> tuple[int, ...]: ...
    @property
    def ndim(self) -> int: ...
    def swapaxes(self, axis1: int, axis2: int, /) -> Self: ...
    def __getitem__(self, key: Any, /) -> Self: ...

type _JustAnyShape = tuple[Never, Never, Never]  # workaround for microsoft/pyright#10232

###

def take_along_axis[ScalarT: np.generic](
    arr: ScalarT | NDArray[ScalarT],
    indices: NDArray[np.integer],
    axis: int | None = -1,
) -> NDArray[ScalarT]: ...

#
def put_along_axis[ScalarT: np.generic](
    arr: NDArray[ScalarT],
    indices: NDArray[np.integer],
    values: ArrayLike,
    axis: int | None,
) -> None: ...

#
@overload
def apply_along_axis[**Tss, ScalarT: np.generic](
    func1d: Callable[Concatenate[np.ndarray, Tss], _ArrayLike[ScalarT]],
    axis: SupportsIndex,
    arr: ArrayLike,
    *args: Tss.args,
    **kwargs: Tss.kwargs,
) -> NDArray[ScalarT]: ...
@overload
def apply_along_axis[**Tss](
    func1d: Callable[Concatenate[np.ndarray, Tss], Any],
    axis: SupportsIndex,
    arr: ArrayLike,
    *args: Tss.args,
    **kwargs: Tss.kwargs,
) -> NDArray[Any]: ...

#
def apply_over_axes[ScalarT: np.generic](
    func: Callable[[np.ndarray, int], NDArray[ScalarT]],
    a: ArrayLike,
    axes: _ShapeLike,
) -> NDArray[ScalarT]: ...

#
@overload  # Nd -> Nd
def expand_dims[ShapeT: _Shape, DTypeT: np.dtype](
    a: np.ndarray[ShapeT, DTypeT],
    axis: tuple[()],
) -> np.ndarray[ShapeT, DTypeT]: ...
@overload  # ?d -> ?d  (workaround)
def expand_dims[DTypeT: np.dtype](
    a: np.ndarray[_JustAnyShape, DTypeT],
    axis: int | tuple[int, ...],
) -> np.ndarray[_AnyShape, DTypeT]: ...
@overload  # 0d -> 1d
def expand_dims[ScalarT: np.generic](
    a: ScalarT | np.ndarray[tuple[()], np.dtype[ScalarT]],
    axis: int | tuple[int],
) -> np.ndarray[tuple[int], np.dtype[ScalarT]]: ...
@overload  # 0d -> 2d
def expand_dims[ScalarT: np.generic](
    a: ScalarT | np.ndarray[tuple[()], np.dtype[ScalarT]],
    axis: tuple[int, int],
) -> np.ndarray[tuple[int, int], np.dtype[ScalarT]]: ...
@overload  # 1d -> 2d
def expand_dims[DTypeT: np.dtype](
    a: np.ndarray[tuple[int], DTypeT],
    axis: int | tuple[int],
) -> np.ndarray[tuple[int, int], DTypeT]: ...
@overload  # 1d -> 3d
def expand_dims[DTypeT: np.dtype](
    a: np.ndarray[tuple[int], DTypeT],
    axis: tuple[int, int],
) -> np.ndarray[tuple[int, int, int], DTypeT]: ...
@overload  # 2d -> 3d
def expand_dims[DTypeT: np.dtype](
    a: np.ndarray[tuple[int, int], DTypeT],
    axis: int | tuple[int],
) -> np.ndarray[tuple[int, int, int], DTypeT]: ...
@overload  # 2d -> 4d
def expand_dims[DTypeT: np.dtype](
    a: np.ndarray[tuple[int, int], DTypeT],
    axis: tuple[int, int],
) -> np.ndarray[tuple[int, int, int, int], DTypeT]: ...
@overload  # Nd -> ?d
def expand_dims[ScalarT: np.generic](
    a: _ArrayLike[ScalarT],
    axis: int | tuple[int, ...],
) -> NDArray[ScalarT]: ...
@overload  # fallback
def expand_dims(a: ArrayLike, axis: int | tuple[int, ...]) -> NDArray[Any]: ...

# keep in sync with `numpy.ma.extras.column_stack`
@overload
def column_stack[ScalarT: np.generic](tup: Sequence[_ArrayLike[ScalarT]]) -> NDArray[ScalarT]: ...
@overload
def column_stack(tup: Sequence[ArrayLike]) -> NDArray[Incomplete]: ...

# keep in sync with `numpy.ma.extras.dstack`
@overload
def dstack[ScalarT: np.generic](tup: Sequence[_ArrayLike[ScalarT]]) -> NDArray[ScalarT]: ...
@overload
def dstack(tup: Sequence[ArrayLike]) -> NDArray[Incomplete]: ...

#
@overload
def array_split[SplitableT: _SupportsSplitOps](
    ary: SplitableT,
    indices_or_sections: _ShapeLike,
    axis: SupportsIndex = 0,
) -> list[SplitableT]: ...
@overload
def array_split[ScalarT: np.generic](
    ary: _ArrayLike[ScalarT],
    indices_or_sections: _ShapeLike,
    axis: SupportsIndex = 0,
) -> list[NDArray[ScalarT]]: ...
@overload
def array_split(ary: ArrayLike, indices_or_sections: _ShapeLike, axis: SupportsIndex = 0) -> list[NDArray[Incomplete]]: ...

#
@overload
def split[SplitableT: _SupportsSplitOps](
    ary: SplitableT,
    indices_or_sections: _ShapeLike,
    axis: SupportsIndex = 0,
) -> list[SplitableT]: ...
@overload
def split[ScalarT: np.generic](
    ary: _ArrayLike[ScalarT],
    indices_or_sections: _ShapeLike,
    axis: SupportsIndex = 0,
) -> list[NDArray[ScalarT]]: ...
@overload
def split(ary: ArrayLike, indices_or_sections: _ShapeLike, axis: SupportsIndex = 0) -> list[NDArray[Incomplete]]: ...

# keep in sync with `numpy.ma.extras.hsplit`
@overload
def hsplit[SplitableT: _SupportsSplitOps](ary: SplitableT, indices_or_sections: _ShapeLike) -> list[SplitableT]: ...
@overload
def hsplit[ScalarT: np.generic](ary: _ArrayLike[ScalarT], indices_or_sections: _ShapeLike) -> list[NDArray[ScalarT]]: ...
@overload
def hsplit(ary: ArrayLike, indices_or_sections: _ShapeLike) -> list[NDArray[Incomplete]]: ...

#
@overload
def vsplit[SplitableT: _SupportsSplitOps](ary: SplitableT, indices_or_sections: _ShapeLike) -> list[SplitableT]: ...
@overload
def vsplit[ScalarT: np.generic](ary: _ArrayLike[ScalarT], indices_or_sections: _ShapeLike) -> list[NDArray[ScalarT]]: ...
@overload
def vsplit(ary: ArrayLike, indices_or_sections: _ShapeLike) -> list[NDArray[Incomplete]]: ...

#
@overload
def dsplit[SplitableT: _SupportsSplitOps](ary: SplitableT, indices_or_sections: _ShapeLike) -> list[SplitableT]: ...
@overload
def dsplit[ScalarT: np.generic](ary: _ArrayLike[ScalarT], indices_or_sections: _ShapeLike) -> list[NDArray[ScalarT]]: ...
@overload
def dsplit(ary: ArrayLike, indices_or_sections: _ShapeLike) -> list[NDArray[Incomplete]]: ...

#
@overload
def kron(a: _ArrayLikeBool_co, b: _ArrayLikeBool_co) -> NDArray[np.bool]: ...
@overload
def kron(a: _ArrayLikeUInt_co, b: _ArrayLikeUInt_co) -> NDArray[np.unsignedinteger]: ...
@overload
def kron(a: _ArrayLikeInt_co, b: _ArrayLikeInt_co) -> NDArray[np.signedinteger]: ...
@overload
def kron(a: _ArrayLikeFloat_co, b: _ArrayLikeFloat_co) -> NDArray[np.floating]: ...
@overload
def kron(a: _ArrayLikeComplex_co, b: _ArrayLikeComplex_co) -> NDArray[np.complexfloating]: ...
@overload
def kron(a: _ArrayLikeObject_co, b: object) -> NDArray[np.object_]: ...
@overload
def kron(a: object, b: _ArrayLikeObject_co) -> NDArray[np.object_]: ...

#
@overload
def tile[ScalarT: np.generic](A: _ArrayLike[ScalarT], reps: _ArrayLikeInt) -> NDArray[ScalarT]: ...
@overload
def tile(A: ArrayLike, reps: _ArrayLikeInt) -> NDArray[Incomplete]: ...



from typing import Any, assert_type

import numpy as np
import numpy.typing as npt

AR_LIKE_b: list[bool]
AR_LIKE_u: list[np.uint32]
AR_LIKE_i: list[int]
AR_LIKE_f: list[float]
AR_LIKE_c: list[complex]
AR_LIKE_U: list[str]
AR_o: npt.NDArray[np.object_]

OUT_f: npt.NDArray[np.float64]

assert_type(np.einsum("i,i->i", AR_LIKE_b, AR_LIKE_b), Any)
assert_type(np.einsum("i,i->i", AR_o, AR_o), Any)
assert_type(np.einsum("i,i->i", AR_LIKE_u, AR_LIKE_u), Any)
assert_type(np.einsum("i,i->i", AR_LIKE_i, AR_LIKE_i), Any)
assert_type(np.einsum("i,i->i", AR_LIKE_f, AR_LIKE_f), Any)
assert_type(np.einsum("i,i->i", AR_LIKE_c, AR_LIKE_c), Any)
assert_type(np.einsum("i,i->i", AR_LIKE_b, AR_LIKE_i), Any)
assert_type(np.einsum("i,i,i,i->i", AR_LIKE_b, AR_LIKE_u, AR_LIKE_i, AR_LIKE_c), Any)

assert_type(np.einsum("i,i->i", AR_LIKE_c, AR_LIKE_c, out=OUT_f), npt.NDArray[np.float64])
assert_type(np.einsum("i,i->i", AR_LIKE_U, AR_LIKE_U, dtype=bool, casting="unsafe", out=OUT_f), npt.NDArray[np.float64])
assert_type(np.einsum("i,i->i", AR_LIKE_f, AR_LIKE_f, dtype="c16"), Any)
assert_type(np.einsum("i,i->i", AR_LIKE_U, AR_LIKE_U, dtype=bool, casting="unsafe"), Any)

assert_type(np.einsum_path("i,i->i", AR_LIKE_b, AR_LIKE_b), tuple[list[Any], str])
assert_type(np.einsum_path("i,i->i", AR_LIKE_u, AR_LIKE_u), tuple[list[Any], str])
assert_type(np.einsum_path("i,i->i", AR_LIKE_i, AR_LIKE_i), tuple[list[Any], str])
assert_type(np.einsum_path("i,i->i", AR_LIKE_f, AR_LIKE_f), tuple[list[Any], str])
assert_type(np.einsum_path("i,i->i", AR_LIKE_c, AR_LIKE_c), tuple[list[Any], str])
assert_type(np.einsum_path("i,i->i", AR_LIKE_b, AR_LIKE_i), tuple[list[Any], str])
assert_type(np.einsum_path("i,i,i,i->i", AR_LIKE_b, AR_LIKE_u, AR_LIKE_i, AR_LIKE_c), tuple[list[Any], str])

assert_type(np.einsum([[1, 1], [1, 1]], AR_LIKE_i, AR_LIKE_i), Any)

assert_type(np.einsum_path([[1, 1], [1, 1]], AR_LIKE_i, AR_LIKE_i), tuple[list[Any], str])

# subscript-free calling convention with float and complex arrays
AR_f: npt.NDArray[np.float64]
AR_c: npt.NDArray[np.complex128]
assert_type(np.einsum(AR_f, [0, 1], AR_f, [1, 0], [0]), Any)
assert_type(np.einsum(AR_c, [0, 1], AR_c, [1, 0], [0]), Any)



"""Tests for the Bayesian beta-binomial simulation helpers."""

from __future__ import annotations

import numpy as np

from abtestwise import bayesian
from abtestwise.bayesian import (
    expected_loss_control,
    expected_loss_treatment,
    simulate_lift_samples,
)


def lift(seed, n=50_000):
    rng = np.random.default_rng(seed)
    return simulate_lift_samples(120, 1000, 145, 1000, 1.0, 1.0, n, rng)


def test_reproducible_with_same_seed():
    a = lift(42)
    b = lift(42)
    assert np.array_equal(a, b)


def test_different_seeds_differ():
    a = lift(1)
    b = lift(2)
    assert not np.array_equal(a, b)


def test_sample_shape():
    samples = lift(0, n=12_345)
    assert samples.shape == (12_345,)


def test_strong_treatment_signal():
    rng = np.random.default_rng(0)
    samples = simulate_lift_samples(100, 1000, 300, 1000, 1.0, 1.0, 50_000, rng)
    assert np.mean(samples) > 0
    assert np.mean(samples > 0) > 0.99
    assert expected_loss_treatment(samples) < 1e-3


def test_identical_arms_are_symmetric():
    rng = np.random.default_rng(7)
    samples = simulate_lift_samples(100, 1000, 100, 1000, 1.0, 1.0, 100_000, rng)
    assert abs(np.mean(samples)) < 0.01
    assert abs(np.mean(samples > 0) - 0.5) < 0.02


def test_expected_loss_formulas_on_known_array():
    samples = np.array([-2.0, -1.0, 1.0, 3.0])
    # treatment loss: mean(max(-lift, 0)) = mean(2, 1, 0, 0) = 0.75
    # control loss:   mean(max(lift, 0))  = mean(0, 0, 1, 3) = 1.0
    assert expected_loss_treatment(samples) == 0.75
    assert expected_loss_control(samples) == 1.0


def test_credible_interval_brackets_median():
    samples = lift(3)
    lower, upper = bayesian.credible_interval_bounds(samples, 0.95)
    median = float(np.median(samples))
    assert lower < median < upper



from typing import Literal, overload

import numpy as np
from numpy.typing import ArrayLike

from matplotlib.tri._triangulation import Triangulation
from matplotlib.tri._triinterpolate import TriInterpolator

class TriRefiner:
    def __init__(self, triangulation: Triangulation) -> None: ...

class UniformTriRefiner(TriRefiner):
    def __init__(self, triangulation: Triangulation) -> None: ...
    @overload
    def refine_triangulation(
        self, *, return_tri_index: Literal[True], subdiv: int = ...
    ) -> tuple[Triangulation, np.ndarray]: ...
    @overload
    def refine_triangulation(
        self, return_tri_index: Literal[False] = ..., subdiv: int = ...
    ) -> Triangulation: ...
    @overload
    def refine_triangulation(
        self, return_tri_index: bool = ..., subdiv: int = ...
    ) -> tuple[Triangulation, np.ndarray] | Triangulation: ...
    def refine_field(
        self,
        z: ArrayLike,
        triinterpolator: TriInterpolator | None = ...,
        subdiv: int = ...,
    ) -> tuple[Triangulation, np.ndarray]: ...



import numpy as np
import pytest
from numpy.testing import assert_allclose
from scipy.signal import whittaker_henderson  # type: ignore[attr-defined]
from scipy.signal._whittaker import (
    _logdet_difference_matrix, _polynomial_fit, _reml, _solveh_banded, _solve_WH_banded,
)
from scipy.stats import special_ortho_group


def _solve_WH_order2_fast(y, lamb):
    """Efficiently solve WH of order 2 according to Weinert.

    Needs order = 2 and n >= 3 (data points).

    Weinert (2007)
    "Efficient computation for Whittaker-Henderson smoothing".
    Computational Statistics and Data Analysis 52:959-74.
    https://doi.org/10.1016/j.csda.2006.11.038
    """
    n = y.shape[0]
    # Convert penalty to convention of Weinert (2007), i.e. A = lambda I + D'D.
    # Note that Weinert denotes the difference matrix M instead of D.
    lamb = 1 / lamb
    # A = LDL' decompositon, i.e. L is unit lower triangular and banded
    # (bandwith=2) and D diagonal.
    # First subdiagonal of L is (-e_1, .., -e_{n-1}) and 2nd subdiagonal
    # (f_1, .., f_{n-2}). Diagonal of D is (d_1, .., d_n).
    # The equation A @ x = lamb * y becomes
    # LD @ b = lamb * y  (I)
    # L.T @ x = b        (II)
    # We shift Weinert's 1-based indices to 0-based indices, Eq. 2.2-2.6
    # Solve problem (I)
    b = np.empty(n)
    e = np.empty(n)
    f = np.empty(n)
    # i=0
    d = 1 + lamb
    f[0] = 1 / d
    mu = 2
    e[0] = mu * f[0]
    b[0] = f[0] * lamb * y[0]
    mu_old = mu
    # i=1
    if n == 3:
        d = 4 + lamb - mu_old * e[0]
        mu = 2 - e[0]
    else:
        d = 5 + lamb - mu_old * e[0]
        mu = 4 - e[0]
    f[1] = 1 / d
    e[1] = mu * f[1]
    b[1] = f[1] * (lamb * y[1] + mu_old * b[0])
    mu_old = mu
    for i in range(2, n-2):
        d = 6 + lamb - mu_old * e[i-1] - f[i-2]
        f[i] = 1 / d
        mu = 4 - e[i-1]
        e[i] = mu * f[i]
        b[i] = f[i] * (lamb * y[i] + mu_old * b[i-1] - b[i-2])
        mu_old = mu
    # i=n-2
    if n >= 4:
        i = n - 2
        d = 5 + lamb - mu_old * e[i-1] - f[i-2]
        f[i] = 1 / d
        mu = 2 - e[i-1]
        e[i] = mu * f[i]
        b[i] = f[i] * (lamb * y[i] + mu_old * b[i-1] - b[i-2])
        mu_old = mu
    # i=n-1
    i = n - 1
    d = 1 + lamb - mu_old * e[i-1] - f[i-2]
    f[i] = 1 / d
    b[i] = f[i] * (lamb * y[i] + mu_old * b[i-1] - b[i-2])

    # Solve problem (II)
    x = np.empty(n)
    x[n-1] = b[n-1]
    x[n-2] = b[n-2] + e[n-2] * x[n-1]
    for i in range(n-3, -1, -1):
        x[i] = b[i] + e[i] * x[i+1] - f[i] * x[i+2]
    return x


def test_solveh_banded():
    n = 10
    # construct positive definite matrix A
    Q = special_ortho_group(n, seed=1234).rvs()
    A = Q @ np.diag(1 + np.arange(n)) @ Q.T
    b = np.arange(n)
    ab = np.zeros((n, n))
    ab[0, :] = np.diagonal(A, 0)
    for i in range(1, n):
        ab[i, :-i] = np.diagonal(A, i)
    x, logdet, _ = _solveh_banded(ab, b, calc_logdet=True)
    assert_allclose(x, np.linalg.solve(A, b), rtol=1e-11)
    assert_allclose(logdet, np.log(np.linalg.det(A)), rtol=1e-11)

    # tridiagonal case
    x, logdet, _ = _solveh_banded(ab[:2], b, calc_logdet=True)
    A_tri = (
        np.diag(np.diag(A)) + np.diag(np.diag(A, -1), -1) + np.diag(np.diag(A, 1), 1)
    )
    assert_allclose(logdet, np.log(np.linalg.det(A_tri)), rtol=1e-11)


@pytest.mark.parametrize(
        ["order", "n"],
        [
            (1, 2), (1, 10),
            (2, 3), (2, 4), (2, 5), (2, 20),
            (5, 6), (5, 7), (5, 10), (5, 20),
            (6, 7), (6, 10), (6, 20),
        ])
def test_logdet_difference_matrix(order, n):
    p = order
    logdet = _logdet_difference_matrix(order=p, n=n)
    D = np.diff(np.eye(n), n=p, axis=0)  # shape (n-p, n)
    assert_allclose(logdet, np.log(np.linalg.det(D @ D.T)), rtol=1e-11)
    eigvals = np.linalg.eigvals(D.T @ D)
    assert_allclose(logdet, np.sum(np.log(eigvals[eigvals > 1e-8])), rtol=1e-11)


@pytest.mark.parametrize("calc_logdet", [False, True])
@pytest.mark.parametrize("lamb", [0.1, 1e20])
def test_polynomial_fit(calc_logdet, lamb):
    """Test that _polynomial_fit works as expected."""
    n = 20
    y = np.sin(2 * np.pi * np.linspace(0, 1, n))
    poly, logdet = _polynomial_fit(y, lamb, calc_logdet=calc_logdet)

    if not calc_logdet:
        assert logdet == 0
    
    poly2, logdet2 = _polynomial_fit(
        y, lamb, weights=1.23 * np.ones(n), calc_logdet=calc_logdet
    )
    assert logdet2 == pytest.approx(logdet, rel=1e-12)
    assert_allclose(poly2, poly)


@pytest.mark.parametrize(
        ["signal", "lamb", "order", "weights", "err", "msg"],
        [
            ([[1, 2, 3] * 3], 1, 2, None, ValueError,
             "Input array signal must be of shape \\(n,\\)"),
            (np.zeros(2), 1, 2, None, ValueError,
             "Input array signal must be at least of shape"),
            (np.arange(10), -0.9, 2, None, ValueError,
             "Parameter lamb must be string"),
            ([1, 2, 3], "XXX", 1, None, ValueError,
             "Parameter lamb must be string 'reml'"),
            ([1, 2, 3], 1, 1.5, None, TypeError,
             "order must be an integer."),
            ([1, 2, 3], 1, 0, None, ValueError,
             "order must be an integer not less than 1"),
            ([1, 2, 3], 1, 2, [0, 1], ValueError,
             "Input array weights must have the same shape as the signal array."),
            ([1, 2, 3], 1, 2, [[0]], ValueError,
             "Input array weights must have the same shape as the signal array."),
            ([1, 2, np.nan], 1, 2, None, ValueError,
             "Input array signal must be finite."),
            ([1, 2, 3], 1, 2, [1, 2, np.nan], ValueError,
             "Input array weights must be finite."),
            ([1, 2, np.nan], 1, 2, [1, 2, 3], ValueError,
             "Input array weights must be zero for all non-finite"),
        ])
def test_whittaker_raises(signal, lamb, order, weights, err, msg):
    """Test that whittaker raises errors."""
    with pytest.raises(err, match=msg):
        whittaker_henderson(signal, lamb=lamb, order=order, weights=weights)


def test_whittaker_small_data():
    """Test that whittaker works on a few data points."""
    # Should work on order + 1 data points. The first 2*order+1 are special.
    la = 1
    y = np.arange(2)
    wh = whittaker_henderson(y, order=1, lamb=la)
    # Analytical solution for order=1 and n=2
    res = (np.array([[1 + la, la], [la, 1 + la]]) / (1 + 2 * la)) @ y
    assert_allclose(wh.x, res, atol=1e-15)
    whittaker_henderson(np.arange(3), order=1, lamb=la)

    y = np.arange(3)
    wh = whittaker_henderson(y, order=2, lamb=la)
    # Analytical solution for order=2 and n=3:
    # import numpy as np
    # from sympy import eye, Symbol
    # n, order = 3, 2
    # la = Symbol("la")
    # D = np.diff(eye(n), n=order, axis=0)
    # A = eye(n) + la * D.T @ D
    # A.inv()
    res = (
        np.array([
            [1 + 5 * la,     2 * la,        -la],
            [    2 * la, 1 + 2 * la,     2 * la],
            [       -la,     2 * la, 1 + 5 * la],
        ]) / (1 + 6 * la)) @ y
    assert_allclose(wh.x, res, atol=1e-15)
    assert wh.lamb == la
    
    y = np.arange(4)
    wh = whittaker_henderson(y, order=2, lamb=la)
    la2 = la ** 2
    res =(np.array([
        [14*la2 + 11*la + 1, 8*la2 + 2*la    , 2*la2 -   la,     -4*la2            ],
        [ 8*la2 +  2*la    , 6*la2 + 7*la + 1, 4*la2 + 4*la,      2*la2 -  la      ],
        [ 2*la2 -    la    , 4*la2 + 4*la    , 6*la2 + 7*la + 1,  8*la2 + 2*la     ],
        [-4*la2            , 2*la2 -   la    , 8*la2 + 2*la    , 14*la2 + 11*la + 1],
    ]) /(20*la2 + 12*la + 1)) @ y
    assert_allclose(wh.x, res, atol=1e-15)
    whittaker_henderson(np.arange(5), order=2, lamb=la)

    whittaker_henderson(np.arange(4), order=3, lamb=la)
    whittaker_henderson(np.arange(5), order=3, lamb=la)
    whittaker_henderson(np.arange(6), order=3, lamb=la)
    whittaker_henderson(np.arange(7), order=3, lamb=la)


@pytest.mark.parametrize("n", [3, 4, 5, 100])
def test_whittaker_direct_vs_fast_order2(n):
    """Test equivalent results"""
    rng = np.random.default_rng(42)
    signal = np.sin(2 * np.pi * np.linspace(0, 1, n)) + rng.standard_normal(n)
    x1, _ = _solve_WH_banded(signal, lamb=1.23, order=2)
    x2 = _solve_WH_order2_fast(signal, lamb=1.23)
    assert_allclose(x1, x2, rtol=1e-11)

    wh = whittaker_henderson(signal, lamb=1.23, order=2)
    assert_allclose(wh.x, x1, rtol=1e-11)


@pytest.mark.parametrize("order", [1, 2, 3, 4])
def test_whittaker_limit_zero_penalty(order):
    """Test whittaker for penalty lamb close to zero."""
    rng = np.random.default_rng(42)
    n = 100
    y = np.sin(2*np.pi * np.linspace(0, 1, n))
    noise = rng.standard_normal(n)
    signal = y + noise

    # In the limit of a zero penalty, the smoothing results in reproducing the signal.
    wh = whittaker_henderson(signal, lamb=1e-7, order=order)
    assert_allclose(wh.x, signal, rtol=1e-4, atol=1e-5)
    assert wh.lamb == 1e-7


@pytest.mark.parametrize("weights", [None, "ones"])
@pytest.mark.parametrize("order", [1, 2, 3, 4])
def test_whittaker_limit_infinity_penalty(weights, order):
    """Test whittaker for penalty lamb close to infinity."""
    n = 100
    y = np.sin(2*np.pi * np.linspace(0, 1, n))
    w = None if weights is None else np.ones_like(y)

    # In the limit of an infinite penalty, the smoothing results in an interpolation
    # polynom of degree = penalty order - 1 (order=2 => linear)
    wh = whittaker_henderson(y, lamb=1e11, order=order, weights=w)
    x_poly = np.arange(len(y))
    poly = np.polynomial.Polynomial.fit(x=x_poly, y=y, deg=order - 1)
    assert_allclose(wh.x, poly(x_poly), rtol=10**(-6 + order), atol=1e-5)
    if order == 2:
        # Linear interpolation:
        # As the sine is positive fom 0 to pi and negative from pi to 2*pi, we expect
        # a negative slope.
        assert np.diff(wh.x)[0] < 0

    if order > 2:
        # Very large lambda should trigger a user warning and produce the exact
        # polynomial fit (separate code path where linear solver breaks down).
        with pytest.warns(
            UserWarning,
            match="The linear solver in Whittaker-Henderson smoothing detected",
        ):
            wh = whittaker_henderson(y, lamb=1e15, order=order, weights=w)
    else:
        wh = whittaker_henderson(y, lamb=1e20, order=order, weights=w)
    assert_allclose(wh.x, poly(x_poly), rtol=1e-10)


def test_whittaker_unpenalized():
    """Test whittaker for lamb=0."""
    n = 10
    y = np.sin(2*np.pi * np.linspace(0, 1, n))
    wh = whittaker_henderson(y, lamb=0)
    assert_allclose(wh.x, y, rtol=1e-11)
    assert not np.may_share_memory(wh.x, y)


def test_whittaker_weights():
    """Test that whittaker with weights of 1 is same as without weights."""
    rng = np.random.default_rng(42)
    n = 100
    y = np.sin(2*np.pi * np.linspace(0, 1, n))
    noise = rng.standard_normal(n)
    signal = y + noise
    wh1 = whittaker_henderson(signal, lamb=1)
    w = np.ones_like(signal)
    wh2 = whittaker_henderson(signal, lamb=1, weights=w)
    assert_allclose(wh1.x, wh2.x, rtol=1e-11)

    # Multiplying penalty and weights by the same number does not change the result.
    wh3 = whittaker_henderson(signal, lamb=3, weights=3*w)
    assert_allclose(wh3.x, wh1.x, rtol=1e-11)


@pytest.mark.parametrize("missing_value", [np.nan, np.inf, 1e-20])
def test_whittaker_zero_weight_interpolation(missing_value):
    """Test that whittaker interpolates where weights are zero."""
    n = 100
    signal = np.sin(2*np.pi * np.linspace(0, 1, n))
    signal[50:] += 2
    w = np.ones_like(signal)
    signal[40:60] = missing_value  # value does not matter
    w[40:60] = 0
    # Note: interpolation is a polynomial of degree = 2 * order - 1.

    # order = 1 => linear interpolation
    wh = whittaker_henderson(signal, lamb=1, order=1, weights=w)
    interp = np.interp(np.arange(40, 60), [39, 60], [wh.x[39], wh.x[60]])
    assert_allclose(wh.x[40:60], interp, rtol=1e-11)

    # order = 2 => cubic interpolation
    wh = whittaker_henderson(signal, lamb=1, order=2, weights=w)
    poly = np.polynomial.Polynomial.fit(
        x=[38, 39, 60, 61], y=[wh.x[38], wh.x[39], wh.x[60], wh.x[61]], deg=3
    )
    assert_allclose(wh.x[40:60], poly(np.arange(40, 60)), rtol=1e-11)


def test_whittaker_zero_weight_extrapolation():
    """Test that whittaker extrapolates where weights are zero."""
    n = 100
    signal = np.sin(2*np.pi * np.linspace(0, 1, n))
    signal[50:] += 2
    w = np.ones_like(signal)
    signal[80:] = np.nan  # value does not matter, np.nan to check arg validation
    w[80:] = 0
    # Note: extrapolation is a polynomial of degree = order - 1.

    # order = 1 => constant extrapolation
    wh = whittaker_henderson(signal, lamb=1, order=1, weights=w)
    assert_allclose(wh.x[80:], wh.x[79], rtol=1e-11)

    # order 2 => linear extrapolation
    wh = whittaker_henderson(signal, lamb=1, order=2, weights=w)
    poly = np.polynomial.Polynomial.fit(x=[78, 79], y=wh.x[78:80], deg=1)
    assert_allclose(wh.x[80:], poly(np.arange(80, 100)), rtol=1e-11)

    # order = 3 => quadratic extrapolation
    wh = whittaker_henderson(signal, lamb=1, order=3, weights=w)
    poly = np.polynomial.Polynomial.fit(x=[77, 78, 79], y=wh.x[77:80], deg=2)
    assert_allclose(wh.x[80:], poly(np.arange(80, 100)), rtol=1e-9)


@pytest.mark.parametrize("order", [1, 2, 3])
def test_reml_criterion(order):
    la = 1.2345
    n = 10
    y = np.sin(2*np.pi * np.linspace(0, 1, n))
    D = np.diff(np.eye(n), n=order, axis=0)
    M = D.T @ D

    def test_reml(la, y):
        # See docstring and comment of _reml.
        A = np.eye(n) + la * M
        x = np.linalg.solve(A, y)
        resid = y - x
        r2 = resid @ resid + la * x @ M @ x
        return -0.5 * ((n-order) * (1 + np.log(r2 / (n - order)))
                       - np.log(np.linalg.det(la * D @ D.T))
                       + np.log(np.linalg.det(A)))

    r1 = _reml(lamb=la, y=y, order=order)
    r2 = test_reml(la=la, y=y)
    assert_allclose(r1, r2, rtol=1e-11)


@pytest.mark.parametrize("weights", [None, "ones"])
def test_whittaker_reml(weights):
    """Test that whittaker with REML criterion works."""
    n = 10
    order = 2
    signal = np.sin(2*np.pi * np.linspace(0, 1, n))
    y = signal.copy()
    # add some (deterministic) noise
    y[::2] += 0.2
    y[1::2] -= 0.2
    w = None if weights is None else np.ones_like(y)

    wh = whittaker_henderson(y, lamb="reml", order=order, weights=w)
    assert wh.lamb > 0

    # Validate against R mgcv
    # library(mgcv)
    # n = 10
    # order = 2
    # signal = sin(2*pi * 0:(n-1) / (n-1))
    # y = signal
    # # add some (deterministic) noise
    # y[seq(1, n, 2)] = y[seq(1, n, 2)] + 0.2
    # y[seq(2, n, 2)] = y[seq(2, n, 2)] - 0.2
    # x <- 1:n
    # b <- gam(y ~ s(x, bs="ps", k=n, m=c(0, 2)), family = gaussian(),
    #          knots = list(x=0:11), method="REML",
    #          control = list(epsilon = 1e-12, mgcv.tol = 1e-12, efs.tol = 1e-12))
    # options(digits=12)
    # b$fitted.values
    expected = [
        0.209959334732, 0.589996568382, 0.938965773497, 0.766582291904, 0.349462653088,
        -0.349462653088, -0.766582291904, -0.938965773497, -0.589996568382,
        -0.209959334732,
    ]
    assert_allclose(wh.x, expected, rtol=1e-7)
    # b$sp
    expected_sp = 5.12904626508
    expected_sp /= 16  # not 100% sure how to justify this 16.
    assert wh.lamb == pytest.approx(expected_sp, rel=1e-7)

    # Double weights should result in double lamb.
    wh2 = whittaker_henderson(y, lamb="reml", order=order, weights=2 * np.ones_like(y))
    assert wh2.lamb == pytest.approx(2 * wh.lamb, rel=1e-7)



"""
Tests for array coercion, mainly through testing `np.array` results directly.
Note that other such tests exist, e.g., in `test_api.py` and many corner-cases
are tested (sometimes indirectly) elsewhere.
"""

from itertools import permutations, product

import pytest
from pytest import param

import numpy as np
import numpy._core._multiarray_umath as ncu
from numpy._core._rational_tests import rational, rational2
from numpy.testing import IS_64BIT, assert_array_equal


def arraylikes():
    """Test parameters for functions converting an array into various array-likes."""

    params = []

    # base array:
    def ndarray(a):
        return a

    params.append(param(ndarray, id="ndarray"))

    # subclass:
    class MyArr(np.ndarray):
        pass

    def subclass(a):
        return a.view(MyArr)

    params.append(subclass)

    class _SequenceLike:
        # Older NumPy versions, sometimes cared whether a protocol array was
        # also _SequenceLike.  This shouldn't matter, but keep it for now
        # for __array__ and not the others.
        def __len__(self):
            raise TypeError

        def __getitem__(self, _, /):
            raise TypeError

    # Array-interface
    class ArrayDunder(_SequenceLike):
        def __init__(self, a):
            self.a = a

        def __array__(self, dtype=None, copy=None):
            if dtype is None:
                return self.a
            return self.a.astype(dtype)

    params.append(param(ArrayDunder, id="__array__"))

    # memory-view
    params.append(param(memoryview, id="memoryview"))

    # Array-interface
    class ArrayInterface:
        def __init__(self, a):
            self.a = a  # need to hold on to keep interface valid
            self.__array_interface__ = a.__array_interface__

    params.append(param(ArrayInterface, id="__array_interface__"))

    # Array-Struct
    class ArrayStruct:
        def __init__(self, a):
            self.a = a  # need to hold on to keep struct valid
            self.__array_struct__ = a.__array_struct__

    params.append(param(ArrayStruct, id="__array_struct__"))

    return params


def scalar_instances(times=True, extended_precision=True, user_dtype=True):
    # Hard-coded list of scalar instances.
    # Floats:
    yield param(np.sqrt(np.float16(5)), id="float16")
    yield param(np.sqrt(np.float32(5)), id="float32")
    yield param(np.sqrt(np.float64(5)), id="float64")
    if extended_precision:
        yield param(np.sqrt(np.longdouble(5)), id="longdouble")

    # Complex:
    yield param(np.sqrt(np.complex64(2 + 3j)), id="complex64")
    yield param(np.sqrt(np.complex128(2 + 3j)), id="complex128")
    if extended_precision:
        yield param(np.sqrt(np.clongdouble(2 + 3j)), id="clongdouble")

    # Bool:
    # XFAIL: Bool should be added, but has some bad properties when it
    # comes to strings, see also gh-9875
    # yield param(np.bool(0), id="bool")

    # Integers:
    yield param(np.int8(2), id="int8")
    yield param(np.int16(2), id="int16")
    yield param(np.int32(2), id="int32")
    yield param(np.int64(2), id="int64")

    yield param(np.uint8(2), id="uint8")
    yield param(np.uint16(2), id="uint16")
    yield param(np.uint32(2), id="uint32")
    yield param(np.uint64(2), id="uint64")

    # Rational:
    if user_dtype:
        yield param(rational(1, 2), id="rational")
        yield param(rational2(1, 2), id="rational2")

    # Cannot create a structured void scalar directly:
    structured = np.array([(1, 3)], "i,i")[0]
    assert isinstance(structured, np.void)
    assert structured.dtype == np.dtype("i,i")
    yield param(structured, id="structured")

    if times:
        # Datetimes and timedelta
        yield param(np.timedelta64(2, "ns"), id="timedelta64[ns]")
        yield param(np.timedelta64(23, "s"), id="timedelta64[s]")
        yield param(np.timedelta64("NaT", "s"), id="timedelta64[s](NaT)")

        yield param(np.datetime64("NaT", "D"), id="datetime64[D](NaT)")
        yield param(np.datetime64("2020-06-07 12:43", "ms"), id="datetime64[ms]")

    # Strings and unstructured void:
    yield param(np.bytes_(b"1234"), id="bytes")
    yield param(np.str_("2345"), id="unicode")
    yield param(np.void(b"4321"), id="unstructured_void")


def is_parametric_dtype(dtype):
    """Returns True if the dtype is a parametric legacy dtype (itemsize
    is 0, or a datetime without units)
    """
    if dtype.itemsize == 0:
        return True
    if issubclass(dtype.type, (np.datetime64, np.timedelta64)):
        if dtype.name.endswith("64"):
            # Generic time units
            return True
    return False


class TestStringDiscovery:
    @pytest.mark.parametrize("obj",
            [object(), 1.2, 10**43, None, "string"],
            ids=["object", "1.2", "10**43", "None", "string"])
    def test_basic_stringlength(self, obj):
        length = len(str(obj))
        expected = np.dtype(f"S{length}")

        assert np.array(obj, dtype="S").dtype == expected
        assert np.array([obj], dtype="S").dtype == expected

        # A nested array is also discovered correctly
        arr = np.array(obj, dtype="O")
        assert np.array(arr, dtype="S").dtype == expected
        # Also if we use the dtype class
        assert np.array(arr, dtype=type(expected)).dtype == expected
        # Check that .astype() behaves identical
        assert arr.astype("S").dtype == expected
        # The DType class is accepted by `.astype()`
        assert arr.astype(type(np.dtype("S"))).dtype == expected

    @pytest.mark.parametrize("obj",
            [object(), 1.2, 10**43, None, "string"],
            ids=["object", "1.2", "10**43", "None", "string"])
    def test_nested_arrays_stringlength(self, obj):
        length = len(str(obj))
        expected = np.dtype(f"S{length}")
        arr = np.array(obj, dtype="O")
        assert np.array([arr, arr], dtype="S").dtype == expected

    @pytest.mark.parametrize("arraylike", arraylikes())
    def test_unpack_first_level(self, arraylike):
        # We unpack exactly one level of array likes
        obj = np.array([None])
        obj[0] = np.array(1.2)
        # the length of the included item, not of the float dtype
        length = len(str(obj[0]))
        expected = np.dtype(f"S{length}")

        obj = arraylike(obj)
        # casting to string usually calls str(obj)
        arr = np.array([obj], dtype="S")
        assert arr.shape == (1, 1)
        assert arr.dtype == expected


class TestScalarDiscovery:
    def test_void_special_case(self):
        # Void dtypes with structures discover tuples as elements
        arr = np.array((1, 2, 3), dtype="i,i,i")
        assert arr.shape == ()
        arr = np.array([(1, 2, 3)], dtype="i,i,i")
        assert arr.shape == (1,)

    def test_char_special_case(self):
        arr = np.array("string", dtype="c")
        assert arr.shape == (6,)
        assert arr.dtype.char == "c"
        arr = np.array(["string"], dtype="c")
        assert arr.shape == (1, 6)
        assert arr.dtype.char == "c"

    def test_char_special_case_deep(self):
        # Check that the character special case errors correctly if the
        # array is too deep:
        nested = ["string"]  # 2 dimensions (due to string being sequence)
        for i in range(ncu.MAXDIMS - 2):
            nested = [nested]

        arr = np.array(nested, dtype='c')
        assert arr.shape == (1,) * (ncu.MAXDIMS - 1) + (6,)
        with pytest.raises(ValueError):
            np.array([nested], dtype="c")

    def test_unknown_object(self):
        arr = np.array(object())
        assert arr.shape == ()
        assert arr.dtype == np.dtype("O")

    @pytest.mark.parametrize("scalar", list(scalar_instances()))
    def test_scalar(self, scalar):
        arr = np.array(scalar)
        assert arr.shape == ()
        assert arr.dtype == scalar.dtype

        arr = np.array([[scalar, scalar]])
        assert arr.shape == (1, 2)
        assert arr.dtype == scalar.dtype

    # Additionally to string this test also runs into a corner case
    # with datetime promotion (the difference is the promotion order).
    @pytest.mark.filterwarnings("ignore:Promotion of numbers:FutureWarning")
    def test_scalar_promotion(self):
        for sc1, sc2 in product(scalar_instances(), scalar_instances()):
            sc1, sc2 = sc1.values[0], sc2.values[0]
            # test all combinations:
            try:
                arr = np.array([sc1, sc2])
            except (TypeError, ValueError):
                # The promotion between two times can fail
                # XFAIL (ValueError): Some object casts are currently undefined
                continue
            assert arr.shape == (2,)
            try:
                dt1, dt2 = sc1.dtype, sc2.dtype
                expected_dtype = np.promote_types(dt1, dt2)
                assert arr.dtype == expected_dtype
            except TypeError as e:
                # Will currently always go to object dtype
                assert arr.dtype == np.dtype("O")

    @pytest.mark.parametrize("scalar", list(scalar_instances()))
    def test_scalar_coercion(self, scalar):
        # This tests various scalar coercion paths, mainly for the numerical
        # types. It includes some paths not directly related to `np.array`.
        if isinstance(scalar, np.inexact):
            # Ensure we have a full-precision number if available
            scalar = type(scalar)((scalar * 2)**0.5)

        # Use casting from object:
        arr = np.array(scalar, dtype=object).astype(scalar.dtype)

        # Test various ways to create an array containing this scalar:
        arr1 = np.array(scalar).reshape(1)
        arr2 = np.array([scalar])
        arr3 = np.empty(1, dtype=scalar.dtype)
        arr3[0] = scalar
        arr4 = np.empty(1, dtype=scalar.dtype)
        arr4[:] = [scalar]
        # All of these methods should yield the same results
        assert_array_equal(arr, arr1)
        assert_array_equal(arr, arr2)
        assert_array_equal(arr, arr3)
        assert_array_equal(arr, arr4)

    @pytest.mark.filterwarnings("ignore::numpy.exceptions.ComplexWarning")
    @pytest.mark.parametrize("cast_to", list(scalar_instances()))
    def test_scalar_coercion_same_as_cast_and_assignment(self, cast_to):
        """
        Test that in most cases:
           * `np.array(scalar, dtype=dtype)`
           * `np.empty((), dtype=dtype)[()] = scalar`
           * `np.array(scalar).astype(dtype)`
        should behave the same.  The only exceptions are parametric dtypes
        (mainly datetime/timedelta without unit) and void without fields.
        """
        dtype = cast_to.dtype  # use to parametrize only the target dtype

        for scalar in scalar_instances(times=False):
            scalar = scalar.values[0]

            if dtype.type == np.void:
                if scalar.dtype.fields is not None and dtype.fields is None:
                    # Here, coercion to "V6" works, but the cast fails.
                    # Since the types are identical, SETITEM takes care of
                    # this, but has different rules than the cast.
                    with pytest.raises(TypeError):
                        np.array(scalar).astype(dtype)
                    np.array(scalar, dtype=dtype)
                    np.array([scalar], dtype=dtype)
                    continue

            # The main test, we first try to use casting and if it succeeds
            # continue below testing that things are the same, otherwise
            # test that the alternative paths at least also fail.
            try:
                cast = np.array(scalar).astype(dtype)
            except (TypeError, ValueError, RuntimeError):
                # coercion should also raise (error type may change)
                with pytest.raises(Exception):  # noqa: B017
                    np.array(scalar, dtype=dtype)

                if (isinstance(scalar, rational) and
                        np.issubdtype(dtype, np.signedinteger)):
                    return

                with pytest.raises(Exception):  # noqa: B017
                    np.array([scalar], dtype=dtype)
                # assignment should also raise
                res = np.zeros((), dtype=dtype)
                with pytest.raises(Exception):  # noqa: B017
                    res[()] = scalar

                return

            # Non error path:
            arr = np.array(scalar, dtype=dtype)
            assert_array_equal(arr, cast)
            # assignment behaves the same
            ass = np.zeros((), dtype=dtype)
            ass[()] = scalar
            assert_array_equal(ass, cast)

    @pytest.mark.parametrize("pyscalar", [10, 10.32, 10.14j, 10**100])
    def test_pyscalar_subclasses(self, pyscalar):
        """NumPy arrays are read/write which means that anything but invariant
        behaviour is on thin ice.  However, we currently are happy to discover
        subclasses of Python float, int, complex the same as the base classes.
        This should potentially be deprecated.
        """
        class MyScalar(type(pyscalar)):
            pass

        res = np.array(MyScalar(pyscalar))
        expected = np.array(pyscalar)
        assert_array_equal(res, expected)

    @pytest.mark.parametrize("dtype_char", np.typecodes["All"])
    def test_default_dtype_instance(self, dtype_char):
        if dtype_char in "SU":
            dtype = np.dtype(dtype_char + "1")
        elif dtype_char == "V":
            # Legacy behaviour was to use V8. The reason was float64 being the
            # default dtype and that having 8 bytes.
            dtype = np.dtype("V8")
        else:
            dtype = np.dtype(dtype_char)

        discovered_dtype, _ = ncu._discover_array_parameters([], type(dtype))

        assert discovered_dtype == dtype
        assert discovered_dtype.itemsize == dtype.itemsize

    @pytest.mark.parametrize("dtype", np.typecodes["Integer"])
    @pytest.mark.parametrize(["scalar", "error"],
            [(np.float64(np.nan), ValueError),
             (np.array(-1).astype(np.ulonglong)[()], OverflowError)])
    def test_scalar_to_int_coerce_does_not_cast(self, dtype, scalar, error):
        """
        Signed integers are currently different in that they do not cast other
        NumPy scalar, but instead use scalar.__int__(). The hardcoded
        exception to this rule is `np.array(scalar, dtype=integer)`.
        """
        dtype = np.dtype(dtype)

        # This is a special case using casting logic. It warns for the NaN
        # but allows the cast (giving undefined behaviour).
        with np.errstate(invalid="ignore"):
            coerced = np.array(scalar, dtype=dtype)
            cast = np.array(scalar).astype(dtype)
        assert_array_equal(coerced, cast)

        # However these fail:
        with pytest.raises(error):
            np.array([scalar], dtype=dtype)
        with pytest.raises(error):
            cast[()] = scalar


class TestTimeScalars:
    @pytest.mark.parametrize("dtype", [np.int64, np.float32])
    @pytest.mark.parametrize("scalar",
            [param(np.timedelta64("NaT", "s"), id="timedelta64[s](NaT)"),
             param(np.timedelta64(123, "s"), id="timedelta64[s]"),
             param(np.datetime64("NaT", "D"), id="datetime64[D](NaT)"),
             param(np.datetime64(1, "D"), id="datetime64[D]")],)
    def test_coercion_basic(self, dtype, scalar):
        # Note the `[scalar]` is there because np.array(scalar) uses stricter
        # `scalar.__int__()` rules for backward compatibility right now.
        arr = np.array(scalar, dtype=dtype)
        cast = np.array(scalar).astype(dtype)
        assert_array_equal(arr, cast)

        ass = np.ones((), dtype=dtype)
        if issubclass(dtype, np.integer):
            with pytest.raises(TypeError):
                # raises, as would np.array([scalar], dtype=dtype), this is
                # conversion from times, but behaviour of integers.
                ass[()] = scalar
        else:
            ass[()] = scalar
            assert_array_equal(ass, cast)

    @pytest.mark.parametrize("dtype", [np.int64, np.float32])
    @pytest.mark.parametrize("value, unit", [param(123, "ns", id="timedelta64[ns]")])
    def test_coercion_timedelta_convert_to_number(self, dtype, value, unit):
        # Only "ns" and "generic" timedeltas can be converted to numbers
        # so these are slightly special.
        scalar = np.timedelta64(value, unit)
        arr = np.array(scalar, dtype=dtype)
        cast = np.array(scalar).astype(dtype)
        ass = np.ones((), dtype=dtype)
        ass[()] = scalar  # raises, as would np.array([scalar], dtype=dtype)

        assert_array_equal(arr, cast)
        assert_array_equal(cast, cast)

    @pytest.mark.parametrize("dtype", [np.int64, np.float32])
    @pytest.mark.parametrize("value, unit",
            [param(12, "generic", id="timedelta64[generic]")])
    def test_coercion_generic_timedelta_convert_to_number(self, dtype, value, unit):
        with pytest.warns(
            DeprecationWarning,
            match="The 'generic' unit for NumPy timedelta is deprecated",
        ):
            scalar = np.timedelta64(value, unit)
            arr = np.array(scalar, dtype=dtype)
            cast = np.array(scalar).astype(dtype)
            ass = np.ones((), dtype=dtype)
            ass[()] = scalar  # raises, as would np.array([scalar], dtype=dtype)

            assert_array_equal(arr, cast)
            assert_array_equal(cast, cast)

    @pytest.mark.parametrize("dtype", ["S6", "U6"])
    @pytest.mark.parametrize(["val", "unit"],
            [param(123, "s", id="[s]"), param(123, "D", id="[D]")])
    def test_coercion_assignment_datetime(self, val, unit, dtype):
        # String from datetime64 assignment is currently special cased to
        # never use casting.  This is because casting will error in this
        # case, and traditionally in most cases the behaviour is maintained
        # like this.  (`np.array(scalar, dtype="U6")` would have failed before)
        # TODO: This discrepancy _should_ be resolved, either by relaxing the
        #       cast, or by deprecating the first part.
        scalar = np.datetime64(val, unit)
        dtype = np.dtype(dtype)
        cut_string = dtype.type(str(scalar)[:6])

        arr = np.array(scalar, dtype=dtype)
        assert arr[()] == cut_string
        ass = np.ones((), dtype=dtype)
        ass[()] = scalar
        assert ass[()] == cut_string

        with pytest.raises(RuntimeError):
            # However, unlike the above assignment using `str(scalar)[:6]`
            # due to being handled by the string DType and not be casting
            # the explicit cast fails:
            np.array(scalar).astype(dtype)

    @pytest.mark.parametrize(["val", "unit"],
            [param(123, "s", id="[s]"), param(123, "D", id="[D]")])
    def test_coercion_assignment_timedelta(self, val, unit):
        scalar = np.timedelta64(val, unit)

        # Unlike datetime64, timedelta allows the unsafe cast:
        np.array(scalar, dtype="S6")
        cast = np.array(scalar).astype("S6")
        ass = np.ones((), dtype="S6")
        ass[()] = scalar
        expected = scalar.astype("S")[:6]
        assert cast[()] == expected
        assert ass[()] == expected

class TestNested:
    def test_nested_simple(self):
        initial = [1.2]
        nested = initial
        for i in range(ncu.MAXDIMS - 1):
            nested = [nested]

        arr = np.array(nested, dtype="float64")
        assert arr.shape == (1,) * ncu.MAXDIMS
        with pytest.raises(ValueError):
            np.array([nested], dtype="float64")

        with pytest.raises(ValueError, match=".*would exceed the maximum"):
            np.array([nested])  # user must ask for `object` explicitly

        arr = np.array([nested], dtype=object)
        assert arr.dtype == np.dtype("O")
        assert arr.shape == (1,) * ncu.MAXDIMS
        assert arr.item() is initial

    def test_pathological_self_containing(self):
        # Test that this also works for two nested sequences
        l = []
        l.append(l)
        arr = np.array([l, l, l], dtype=object)
        assert arr.shape == (3,) + (1,) * (ncu.MAXDIMS - 1)

        # Also check a ragged case:
        arr = np.array([l, [None], l], dtype=object)
        assert arr.shape == (3, 1)

    @pytest.mark.parametrize("arraylike", arraylikes())
    def test_nested_arraylikes(self, arraylike):
        # We try storing an array like into an array, but the array-like
        # will have too many dimensions.  This means the shape discovery
        # decides that the array-like must be treated as an object (a special
        # case of ragged discovery).  The result will be an array with one
        # dimension less than the maximum dimensions, and the array being
        # assigned to it (which does work for object or if `float(arraylike)`
        # works).
        initial = arraylike(np.ones((1, 1)))

        nested = initial
        for i in range(ncu.MAXDIMS - 1):
            nested = [nested]

        with pytest.raises(ValueError, match=".*would exceed the maximum"):
            # It will refuse to assign the array into
            np.array(nested, dtype="float64")

        # If this is object, we end up assigning a (1, 1) array into (1,)
        # (due to running out of dimensions), this is currently supported but
        # a special case which is not ideal.
        arr = np.array(nested, dtype=object)
        assert arr.shape == (1,) * ncu.MAXDIMS
        assert arr.item() == np.array(initial).item()

    @pytest.mark.parametrize("arraylike", arraylikes())
    def test_uneven_depth_ragged(self, arraylike):
        arr = np.arange(4).reshape((2, 2))
        arr = arraylike(arr)

        # Array is ragged in the second dimension already:
        out = np.array([arr, [arr]], dtype=object)
        assert out.shape == (2,)
        assert out[0] is arr
        assert type(out[1]) is list

        # Array is ragged in the third dimension:
        with pytest.raises(ValueError):
            # This is a broadcast error during assignment, because
            # the array shape would be (2, 2, 2) but `arr[0, 0] = arr` fails.
            np.array([arr, [arr, arr]], dtype=object)

    def test_empty_sequence(self):
        arr = np.array([[], [1], [[1]]], dtype=object)
        assert arr.shape == (3,)

        # The empty sequence stops further dimension discovery, so the
        # result shape will be (0,) which leads to an error during:
        with pytest.raises(ValueError):
            np.array([[], np.empty((0, 1))], dtype=object)

    def test_array_of_different_depths(self):
        # When multiple arrays (or array-likes) are included in a
        # sequences and have different depth, we currently discover
        # as many dimensions as they share. (see also gh-17224)
        arr = np.zeros((3, 2))
        mismatch_first_dim = np.zeros((1, 2))
        mismatch_second_dim = np.zeros((3, 3))

        dtype, shape = ncu._discover_array_parameters(
            [arr, mismatch_second_dim], dtype=np.dtype("O"))
        assert shape == (2, 3)

        dtype, shape = ncu._discover_array_parameters(
            [arr, mismatch_first_dim], dtype=np.dtype("O"))
        assert shape == (2,)
        # The second case is currently supported because the arrays
        # can be stored as objects:
        res = np.asarray([arr, mismatch_first_dim], dtype=np.dtype("O"))
        assert res[0] is arr
        assert res[1] is mismatch_first_dim


class TestBadSequences:
    # These are tests for bad objects passed into `np.array`, in general
    # these have undefined behaviour.  In the old code they partially worked
    # when now they will fail.  We could (and maybe should) create a copy
    # of all sequences to be safe against bad-actors.

    def test_growing_list(self):
        # List to coerce, `mylist` will append to it during coercion
        obj = []

        class mylist(list):
            def __len__(self):
                obj.append([1, 2])
                return super().__len__()

        obj.append(mylist([1, 2]))

        with pytest.raises(RuntimeError):
            np.array(obj)

    # Note: We do not test a shrinking list.  These do very evil things
    #       and the only way to fix them would be to copy all sequences.
    #       (which may be a real option in the future).

    def test_mutated_list(self):
        # List to coerce, `mylist` will mutate the first element
        obj = []

        class mylist(list):
            def __len__(self):
                obj[0] = [2, 3]  # replace with a different list.
                return super().__len__()

        obj.append([2, 3])
        obj.append(mylist([1, 2]))
        # Does not crash:
        np.array(obj)

    def test_replace_0d_array(self):
        # List to coerce, `mylist` will mutate the first element
        obj = []

        class baditem:
            def __len__(self):
                obj[0][0] = 2  # replace with a different list.
                raise ValueError("not actually a sequence!")

            def __getitem__(self, _, /):
                pass

        # Runs into a corner case in the new code, the `array(2)` is cached
        # so replacing it invalidates the cache.
        obj.append([np.array(2), baditem()])
        with pytest.raises(RuntimeError):
            np.array(obj)


class TestArrayLikes:
    @pytest.mark.parametrize("arraylike", arraylikes())
    def test_0d_object_special_case(self, arraylike):
        arr = np.array(0.)
        obj = arraylike(arr)
        # A single array-like is always converted:
        res = np.array(obj, dtype=object)
        assert_array_equal(arr, res)

        # But a single 0-D nested array-like never:
        res = np.array([obj], dtype=object)
        assert res[0] is obj

    @pytest.mark.parametrize("arraylike", arraylikes())
    @pytest.mark.parametrize("arr", [np.array(0.), np.arange(4)])
    def test_object_assignment_special_case(self, arraylike, arr):
        obj = arraylike(arr)
        empty = np.arange(1, dtype=object)
        empty[:] = [obj]
        assert empty[0] is obj

    def test_0d_generic_special_case(self):
        class ArraySubclass(np.ndarray):
            def __float__(self):
                raise TypeError("e.g. quantities raise on this")

        arr = np.array(0.)
        obj = arr.view(ArraySubclass)
        res = np.array(obj)
        # The subclass is simply cast:
        assert_array_equal(arr, res)

        # If the 0-D array-like is included, __float__ is currently
        # guaranteed to be used.  We may want to change that, quantities
        # and masked arrays half make use of this.
        with pytest.raises(TypeError):
            np.array([obj])

        # The same holds for memoryview:
        obj = memoryview(arr)
        res = np.array(obj)
        assert_array_equal(arr, res)
        with pytest.raises(ValueError):
            # The error type does not matter much here.
            np.array([obj])

    def test_arraylike_classes(self):
        # The classes of array-likes should generally be acceptable to be
        # stored inside a numpy (object) array.  This tests all of the
        # special attributes (since all are checked during coercion).
        arr = np.array(np.int64)
        assert arr[()] is np.int64
        arr = np.array([np.int64])
        assert arr[0] is np.int64

        # This also works for properties/unbound methods:
        class ArrayLike:
            @property
            def __array_interface__(self):
                pass

            @property
            def __array_struct__(self):
                pass

            def __array__(self, dtype=None, copy=None):
                pass

        arr = np.array(ArrayLike)
        assert arr[()] is ArrayLike
        arr = np.array([ArrayLike])
        assert arr[0] is ArrayLike

    @pytest.mark.skipif(not IS_64BIT, reason="Needs 64bit platform")
    @pytest.mark.thread_unsafe(reason="large slow test in parallel")
    def test_too_large_array_error_paths(self):
        """Test the error paths, including for memory leaks"""
        arr = np.array(0, dtype="uint8")
        # Guarantees that a contiguous copy won't work:
        arr = np.broadcast_to(arr, 2**62)

        for i in range(5):
            # repeat, to ensure caching cannot have an effect:
            with pytest.raises(MemoryError):
                np.array(arr)
            with pytest.raises(MemoryError):
                np.array([arr])

    @pytest.mark.parametrize("attribute",
        ["__array_interface__", "__array__", "__array_struct__"])
    @pytest.mark.parametrize("error", [RecursionError, MemoryError])
    def test_bad_array_like_attributes(self, attribute, error):
        # RecursionError and MemoryError are considered fatal. All errors
        # (except AttributeError) should probably be raised in the future,
        # but shapely made use of it, so it will require a deprecation.

        class BadInterface:
            def __getattr__(self, attr):
                if attr == attribute:
                    raise error
                super().__getattr__(attr)

        with pytest.raises(error):
            np.array(BadInterface())

    @pytest.mark.parametrize("error", [RecursionError, MemoryError])
    def test_bad_array_like_bad_length(self, error):
        # RecursionError and MemoryError are considered "critical" in
        # sequences. We could expand this more generally though. (NumPy 1.20)
        class BadSequence:
            def __len__(self):
                raise error

            def __getitem__(self, _, /):
                # must have getitem to be a Sequence
                return 1

        with pytest.raises(error):
            np.array(BadSequence())

    def test_array_interface_descr_optional(self):
        # The descr should be optional regression test for gh-27249
        arr = np.ones(10, dtype="V10")
        iface = arr.__array_interface__
        iface.pop("descr")

        class MyClass:
            __array_interface__ = iface

        assert_array_equal(np.asarray(MyClass), arr)


class TestAsArray:
    """Test expected behaviors of ``asarray``."""

    def test_dtype_identity(self):
        """Confirm the intended behavior for *dtype* kwarg.

        The result of ``asarray()`` should have the dtype provided through the
        keyword argument, when used. This forces unique array handles to be
        produced for unique np.dtype objects, but (for equivalent dtypes), the
        underlying data (the base object) is shared with the original array
        object.

        Ref https://github.com/numpy/numpy/issues/1468
        """
        int_array = np.array([1, 2, 3], dtype='i')
        assert np.asarray(int_array) is int_array

        # The character code resolves to the singleton dtype object provided
        # by the numpy package.
        assert np.asarray(int_array, dtype='i') is int_array

        # Derive a dtype from n.dtype('i'), but add a metadata object to force
        # the dtype to be distinct.
        unequal_type = np.dtype('i', metadata={'spam': True})
        annotated_int_array = np.asarray(int_array, dtype=unequal_type)
        assert annotated_int_array is not int_array
        assert annotated_int_array.base is int_array
        # Create an equivalent descriptor with a new and distinct dtype
        # instance.
        equivalent_requirement = np.dtype('i', metadata={'spam': True})
        annotated_int_array_alt = np.asarray(annotated_int_array,
                                             dtype=equivalent_requirement)
        assert unequal_type == equivalent_requirement
        assert unequal_type is not equivalent_requirement
        assert annotated_int_array_alt is not annotated_int_array
        assert annotated_int_array_alt.dtype is equivalent_requirement

        # Check the same logic for a pair of C types whose equivalence may vary
        # between computing environments.
        # Find an equivalent pair.
        integer_type_codes = ('i', 'l', 'q')
        integer_dtypes = [np.dtype(code) for code in integer_type_codes]
        typeA = None
        typeB = None
        for typeA, typeB in permutations(integer_dtypes, r=2):
            if typeA == typeB:
                assert typeA is not typeB
                break
        assert isinstance(typeA, np.dtype) and isinstance(typeB, np.dtype)

        # These ``asarray()`` calls may produce a new view or a copy,
        # but never the same object.
        long_int_array = np.asarray(int_array, dtype='l')
        long_long_int_array = np.asarray(int_array, dtype='q')
        assert long_int_array is not int_array
        assert long_long_int_array is not int_array
        assert np.asarray(long_int_array, dtype='q') is not long_int_array
        array_a = np.asarray(int_array, dtype=typeA)
        assert typeA == typeB
        assert typeA is not typeB
        assert array_a.dtype is typeA
        assert array_a is not np.asarray(array_a, dtype=typeB)
        assert np.asarray(array_a, dtype=typeB).dtype is typeB
        assert array_a is np.asarray(array_a, dtype=typeB).base


class TestSpecialAttributeLookupFailure:
    # An exception was raised while fetching the attribute

    class WeirdArrayLike:
        @property
        def __array__(self, dtype=None, copy=None):
            raise RuntimeError("oops!")

    class WeirdArrayInterface:
        @property
        def __array_interface__(self):
            raise RuntimeError("oops!")

    def test_deprecated(self):
        with pytest.raises(RuntimeError):
            np.array(self.WeirdArrayLike())
        with pytest.raises(RuntimeError):
            np.array(self.WeirdArrayInterface())


def test_subarray_from_array_construction():
    # Arrays are more complex, since they "broadcast" on success:
    arr = np.array([1, 2])

    res = arr.astype("2i")
    assert_array_equal(res, [[1, 1], [2, 2]])

    res = np.array(arr, dtype="(2,)i")

    assert_array_equal(res, [[1, 1], [2, 2]])

    res = np.array([[(1,), (2,)], arr], dtype="2i")
    assert_array_equal(res, [[[1, 1], [2, 2]], [[1, 1], [2, 2]]])

    # Also try a multi-dimensional example:
    arr = np.arange(5 * 2).reshape(5, 2)
    expected = np.broadcast_to(arr[:, :, np.newaxis, np.newaxis], (5, 2, 2, 2))

    res = arr.astype("(2,2)f")
    assert_array_equal(res, expected)

    res = np.array(arr, dtype="(2,2)f")
    assert_array_equal(res, expected)


def test_empty_string():
    # Empty strings are unfortunately often converted to S1 and we need to
    # make sure we are filling the S1 and not the (possibly) detected S0
    # result.  This should likely just return S0 and if not maybe the decision
    # to return S1 should be moved.
    res = np.array([""] * 10, dtype="S")
    assert_array_equal(res, np.array("\0", "S1"))
    assert res.dtype == "S1"

    arr = np.array([""] * 10, dtype=object)

    res = arr.astype("S")
    assert_array_equal(res, b"")
    assert res.dtype == "S1"

    res = np.array(arr, dtype="S")
    assert_array_equal(res, b"")
    # TODO: This is arguably weird/wrong, but seems old:
    assert res.dtype == f"S{np.dtype('O').itemsize}"

    res = np.array([[""] * 10, arr], dtype="S")
    assert_array_equal(res, b"")
    assert res.shape == (2, 10)
    assert res.dtype == "S1"


@pytest.mark.parametrize("dtype", ["S", "U", object])
@pytest.mark.parametrize("res_dt,hug_val",
    [("float16", "1e30"), ("float32", "1e200")])
def test_string_to_float_coercion_errors(dtype, res_dt, hug_val):
    # This test primarily tests setitem
    val = np.array(["3M"], dtype=dtype)[0]  # use the scalar

    with pytest.raises(ValueError):
        np.array(val, dtype=res_dt)

    val = np.array([hug_val], dtype=dtype)[0]  # use the scalar

    with np.errstate(all="warn"):
        with pytest.warns(RuntimeWarning):
            np.array(val, dtype=res_dt)

    with np.errstate(all="raise"):
        with pytest.raises(FloatingPointError):
            np.array(val, dtype=res_dt)



"""
Routines to adjust subplot params so that subplots are
nicely fit in the figure. In doing so, only axis labels, tick labels, Axes
titles and offsetboxes that are anchored to Axes are currently considered.

Internally, this module assumes that the margins (left margin, etc.) which are
differences between ``Axes.get_tightbbox`` and ``Axes.bbox`` are independent of
Axes position. This may fail if ``Axes.adjustable`` is ``datalim`` as well as
such cases as when left or right margin are affected by xlabel.
"""

import numpy as np

import matplotlib as mpl
from matplotlib import _api, artist as martist
from matplotlib.font_manager import FontProperties
from matplotlib.transforms import Bbox


def _auto_adjust_subplotpars(
        fig, renderer, shape, span_pairs, subplot_list,
        ax_bbox_list=None, pad=1.08, h_pad=None, w_pad=None, rect=None):
    """
    Return a dict of subplot parameters to adjust spacing between subplots
    or ``None`` if resulting Axes would have zero height or width.

    Note that this function ignores geometry information of subplot itself, but
    uses what is given by the *shape* and *subplot_list* parameters.  Also, the
    results could be incorrect if some subplots have ``adjustable=datalim``.

    Parameters
    ----------
    shape : tuple[int, int]
        Number of rows and columns of the grid.
    span_pairs : list[tuple[slice, slice]]
        List of rowspans and colspans occupied by each subplot.
    subplot_list : list of subplots
        List of subplots that will be used to calculate optimal subplot_params.
    pad : float
        Padding between the figure edge and the edges of subplots, as a
        fraction of the font size.
    h_pad, w_pad : float
        Padding (height/width) between edges of adjacent subplots, as a
        fraction of the font size.  Defaults to *pad*.
    rect : tuple
        (left, bottom, right, top), default: None.
    """
    rows, cols = shape

    font_size_inch = (FontProperties(
        size=mpl.rcParams["font.size"]).get_size_in_points() / 72)
    pad_inch = pad * font_size_inch
    vpad_inch = h_pad * font_size_inch if h_pad is not None else pad_inch
    hpad_inch = w_pad * font_size_inch if w_pad is not None else pad_inch

    if len(span_pairs) != len(subplot_list) or len(subplot_list) == 0:
        raise ValueError

    if rect is None:
        margin_left = margin_bottom = margin_right = margin_top = None
    else:
        margin_left, margin_bottom, _right, _top = rect
        margin_right = 1 - _right if _right else None
        margin_top = 1 - _top if _top else None

    vspaces = np.zeros((rows + 1, cols))
    hspaces = np.zeros((rows, cols + 1))

    if ax_bbox_list is None:
        ax_bbox_list = [
            Bbox.union([ax.get_position(original=True) for ax in subplots])
            for subplots in subplot_list]

    for subplots, ax_bbox, (rowspan, colspan) in zip(
            subplot_list, ax_bbox_list, span_pairs):
        if all(not ax.get_visible() for ax in subplots):
            continue

        bb = []
        for ax in subplots:
            if ax.get_visible():
                bb += [martist._get_tightbbox_for_layout_only(ax, renderer)]

        tight_bbox_raw = Bbox.union(bb)
        tight_bbox = fig.transFigure.inverted().transform_bbox(tight_bbox_raw)

        hspaces[rowspan, colspan.start] += ax_bbox.xmin - tight_bbox.xmin  # l
        hspaces[rowspan, colspan.stop] += tight_bbox.xmax - ax_bbox.xmax  # r
        vspaces[rowspan.start, colspan] += tight_bbox.ymax - ax_bbox.ymax  # t
        vspaces[rowspan.stop, colspan] += ax_bbox.ymin - tight_bbox.ymin  # b

    fig_width_inch, fig_height_inch = fig.get_size_inches()

    # margins can be negative for Axes with aspect applied, so use max(, 0) to
    # make them nonnegative.
    if not margin_left:
        margin_left = max(hspaces[:, 0].max(), 0) + pad_inch/fig_width_inch
        suplabel = fig._supylabel
        if suplabel and suplabel.get_in_layout():
            rel_width = fig.transFigure.inverted().transform_bbox(
                suplabel.get_window_extent(renderer)).width
            margin_left += rel_width + pad_inch/fig_width_inch
    if not margin_right:
        margin_right = max(hspaces[:, -1].max(), 0) + pad_inch/fig_width_inch
    if not margin_top:
        margin_top = max(vspaces[0, :].max(), 0) + pad_inch/fig_height_inch
        if fig._suptitle and fig._suptitle.get_in_layout():
            rel_height = fig.transFigure.inverted().transform_bbox(
                fig._suptitle.get_window_extent(renderer)).height
            margin_top += rel_height + pad_inch/fig_height_inch
    if not margin_bottom:
        margin_bottom = max(vspaces[-1, :].max(), 0) + pad_inch/fig_height_inch
        suplabel = fig._supxlabel
        if suplabel and suplabel.get_in_layout():
            rel_height = fig.transFigure.inverted().transform_bbox(
                suplabel.get_window_extent(renderer)).height
            margin_bottom += rel_height + pad_inch/fig_height_inch

    if margin_left + margin_right >= 1:
        _api.warn_external('Tight layout not applied. The left and right '
                           'margins cannot be made large enough to '
                           'accommodate all Axes decorations.')
        return None
    if margin_bottom + margin_top >= 1:
        _api.warn_external('Tight layout not applied. The bottom and top '
                           'margins cannot be made large enough to '
                           'accommodate all Axes decorations.')
        return None

    kwargs = dict(left=margin_left,
                  right=1 - margin_right,
                  bottom=margin_bottom,
                  top=1 - margin_top)

    if cols > 1:
        hspace = hspaces[:, 1:-1].max() + hpad_inch / fig_width_inch
        # axes widths:
        h_axes = (1 - margin_right - margin_left - hspace * (cols - 1)) / cols
        if h_axes < 0:
            _api.warn_external('Tight layout not applied. tight_layout '
                               'cannot make Axes width small enough to '
                               'accommodate all Axes decorations')
            return None
        else:
            kwargs["wspace"] = hspace / h_axes
    if rows > 1:
        vspace = vspaces[1:-1, :].max() + vpad_inch / fig_height_inch
        v_axes = (1 - margin_top - margin_bottom - vspace * (rows - 1)) / rows
        if v_axes < 0:
            _api.warn_external('Tight layout not applied. tight_layout '
                               'cannot make Axes height small enough to '
                               'accommodate all Axes decorations.')
            return None
        else:
            kwargs["hspace"] = vspace / v_axes

    return kwargs


def get_subplotspec_list(axes_list, grid_spec=None):
    """
    Return a list of subplotspec from the given list of Axes.

    For an instance of Axes that does not support subplotspec, None is inserted
    in the list.

    If grid_spec is given, None is inserted for those not from the given
    grid_spec.
    """
    subplotspec_list = []
    for ax in axes_list:
        axes_or_locator = ax.get_axes_locator()
        if axes_or_locator is None:
            axes_or_locator = ax

        if hasattr(axes_or_locator, "get_subplotspec"):
            subplotspec = axes_or_locator.get_subplotspec()
            if subplotspec is not None:
                subplotspec = subplotspec.get_topmost_subplotspec()
                gs = subplotspec.get_gridspec()
                if grid_spec is not None:
                    if gs != grid_spec:
                        subplotspec = None
                elif gs.locally_modified_subplot_params():
                    subplotspec = None
        else:
            subplotspec = None

        subplotspec_list.append(subplotspec)

    return subplotspec_list


def get_tight_layout_figure(fig, axes_list, subplotspec_list, renderer,
                            pad=1.08, h_pad=None, w_pad=None, rect=None):
    """
    Return subplot parameters for tight-layouted-figure with specified padding.

    Parameters
    ----------
    fig : Figure
    axes_list : list of Axes
    subplotspec_list : list of `.SubplotSpec`
        The subplotspecs of each Axes.
    renderer : renderer
    pad : float
        Padding between the figure edge and the edges of subplots, as a
        fraction of the font size.
    h_pad, w_pad : float
        Padding (height/width) between edges of adjacent subplots.  Defaults to
        *pad*.
    rect : tuple (left, bottom, right, top), default: None.
        rectangle in normalized figure coordinates
        that the whole subplots area (including labels) will fit into.
        Defaults to using the entire figure.

    Returns
    -------
    subplotspec or None
        subplotspec kwargs to be passed to `.Figure.subplots_adjust` or
        None if tight_layout could not be accomplished.
    """

    # Multiple Axes can share same subplotspec (e.g., if using axes_grid1);
    # we need to group them together.
    ss_to_subplots = {ss: [] for ss in subplotspec_list}
    for ax, ss in zip(axes_list, subplotspec_list):
        ss_to_subplots[ss].append(ax)
    if ss_to_subplots.pop(None, None):
        _api.warn_external(
            "This figure includes Axes that are not compatible with "
            "tight_layout, so results might be incorrect.")
    if not ss_to_subplots:
        return {}
    subplot_list = list(ss_to_subplots.values())
    ax_bbox_list = [ss.get_position(fig) for ss in ss_to_subplots]

    max_nrows = max(ss.get_gridspec().nrows for ss in ss_to_subplots)
    max_ncols = max(ss.get_gridspec().ncols for ss in ss_to_subplots)

    span_pairs = []
    for ss in ss_to_subplots:
        # The intent here is to support Axes from different gridspecs where
        # one's nrows (or ncols) is a multiple of the other (e.g. 2 and 4),
        # but this doesn't actually work because the computed wspace, in
        # relative-axes-height, corresponds to different physical spacings for
        # the 2-row grid and the 4-row grid.  Still, this code is left, mostly
        # for backcompat.
        rows, cols = ss.get_gridspec().get_geometry()
        div_row, mod_row = divmod(max_nrows, rows)
        div_col, mod_col = divmod(max_ncols, cols)
        if mod_row != 0:
            _api.warn_external('tight_layout not applied: number of rows '
                               'in subplot specifications must be '
                               'multiples of one another.')
            return {}
        if mod_col != 0:
            _api.warn_external('tight_layout not applied: number of '
                               'columns in subplot specifications must be '
                               'multiples of one another.')
            return {}
        span_pairs.append((
            slice(ss.rowspan.start * div_row, ss.rowspan.stop * div_row),
            slice(ss.colspan.start * div_col, ss.colspan.stop * div_col)))

    kwargs = _auto_adjust_subplotpars(fig, renderer,
                                      shape=(max_nrows, max_ncols),
                                      span_pairs=span_pairs,
                                      subplot_list=subplot_list,
                                      ax_bbox_list=ax_bbox_list,
                                      pad=pad, h_pad=h_pad, w_pad=w_pad)

    # kwargs can be none if tight_layout fails...
    if rect is not None and kwargs is not None:
        # if rect is given, the whole subplots area (including
        # labels) will fit into the rect instead of the
        # figure. Note that the rect argument of
        # *auto_adjust_subplotpars* specify the area that will be
        # covered by the total area of axes.bbox. Thus we call
        # auto_adjust_subplotpars twice, where the second run
        # with adjusted rect parameters.

        left, bottom, right, top = rect
        if left is not None:
            left += kwargs["left"]
        if bottom is not None:
            bottom += kwargs["bottom"]
        if right is not None:
            right -= (1 - kwargs["right"])
        if top is not None:
            top -= (1 - kwargs["top"])

        kwargs = _auto_adjust_subplotpars(fig, renderer,
                                          shape=(max_nrows, max_ncols),
                                          span_pairs=span_pairs,
                                          subplot_list=subplot_list,
                                          ax_bbox_list=ax_bbox_list,
                                          pad=pad, h_pad=h_pad, w_pad=w_pad,
                                          rect=(left, bottom, right, top))

    return kwargs



""" Sketching-based Matrix Computations """

# Author: Jordi Montes <jomsdev@gmail.com>
# August 28, 2017

import numpy as np

from scipy._lib._util import (check_random_state, rng_integers,
                              _transition_to_rng, _apply_over_batch)

__all__ = ['clarkson_woodruff_transform']


def cwt_matrix(n_rows, n_columns, rng=None):
    r"""
    Generate a matrix S which represents a Clarkson-Woodruff transform.

    Given the desired size of matrix, the method returns a matrix S of size
    (n_rows, n_columns) where each column has all the entries set to 0
    except for one position which has been randomly set to +1 or -1 with
    equal probability.

    Parameters
    ----------
    n_rows : int
        Number of rows of S
    n_columns : int
        Number of columns of S
    rng : `numpy.random.Generator`, optional
        Pseudorandom number generator state. When `rng` is None, a new
        `numpy.random.Generator` is created using entropy from the
        operating system. Types other than `numpy.random.Generator` are
        passed to `numpy.random.default_rng` to instantiate a ``Generator``.

    Returns
    -------
    S : (n_rows, n_columns) csc_array
        The returned matrix has ``n_columns`` nonzero entries.

    Notes
    -----
    Given a matrix A, with probability at least 9/10,
    .. math:: \|SA\| = (1 \pm \epsilon)\|A\|
    Where the error epsilon is related to the size of S.
    """
    # lazy import to prevent to prevent sparse dependency for whole module (gh-23420)
    from scipy.sparse import csc_array
    rng = check_random_state(rng)
    rows = rng_integers(rng, 0, n_rows, n_columns)
    cols = np.arange(n_columns+1)
    signs = rng.choice([1, -1], n_columns)
    S = csc_array((signs, rows, cols), shape=(n_rows, n_columns))
    return S


@_transition_to_rng("seed", position_num=2)
def clarkson_woodruff_transform(input_matrix, sketch_size, rng=None):
    r"""
    Applies a Clarkson-Woodruff Transform/sketch to the input matrix.

    Given an input_matrix ``A`` of size ``(n, d)``, compute a matrix ``A'`` of
    size (sketch_size, d) so that

    .. math:: \|Ax\| \approx \|A'x\|

    with high probability via the Clarkson-Woodruff Transform, otherwise
    known as the CountSketch matrix.

    The documentation is written assuming array arguments are of specified
    "core" shapes. However, array argument(s) of this function may have additional
    "batch" dimensions prepended to the core shape. In this case, the array is treated
    as a batch of lower-dimensional slices; see :ref:`linalg_batch` for details.

    Parameters
    ----------
    input_matrix : array_like, shape (..., n, d)
        Input matrix.
    sketch_size : int
        Number of rows for the sketch.
    rng : `numpy.random.Generator`, optional
        Pseudorandom number generator state. When `rng` is None, a new
        `numpy.random.Generator` is created using entropy from the
        operating system. Types other than `numpy.random.Generator` are
        passed to `numpy.random.default_rng` to instantiate a ``Generator``.

    Returns
    -------
    A' : array_like
        Sketch of the input matrix ``A``, of size ``(sketch_size, d)``.

    Notes
    -----
    To make the statement

    .. math:: \|Ax\| \approx \|A'x\|

    precise, observe the following result which is adapted from the
    proof of Theorem 14 of [2]_ via Markov's Inequality. If we have
    a sketch size ``sketch_size=k`` which is at least

    .. math:: k \geq \frac{2}{\epsilon^2\delta}

    Then for any fixed vector ``x``,

    .. math:: \|Ax\| = (1\pm\epsilon)\|A'x\|

    with probability at least one minus delta.

    This implementation takes advantage of sparsity: computing
    a sketch takes time proportional to ``A.nnz``. Data ``A`` which
    is in ``scipy.sparse.csc_array`` format gives the quickest
    computation time for sparse input.

    >>> import numpy as np
    >>> from scipy import linalg
    >>> from scipy import sparse
    >>> rng = np.random.default_rng()
    >>> *shape, density, sketch_n_rows = 15000, 100, 0.01, 200
    >>> A = sparse.random_array(shape, density=density, format='csc')
    >>> B = sparse.random_array(shape, density=density, format='csr')
    >>> C = sparse.random_array(shape, density=density, format='coo')
    >>> D = rng.standard_normal(shape)
    >>> SA = linalg.clarkson_woodruff_transform(A, sketch_n_rows) # fastest
    >>> SB = linalg.clarkson_woodruff_transform(B, sketch_n_rows) # fast
    >>> SC = linalg.clarkson_woodruff_transform(C, sketch_n_rows) # slower
    >>> SD = linalg.clarkson_woodruff_transform(D, sketch_n_rows) # slowest

    That said, this method does perform well on dense inputs, just slower
    on a relative scale.

    References
    ----------
    .. [1] Kenneth L. Clarkson and David P. Woodruff. Low rank approximation
           and regression in input sparsity time. In STOC, 2013.
    .. [2] David P. Woodruff. Sketching as a tool for numerical linear algebra.
           In Foundations and Trends in Theoretical Computer Science, 2014.

    Examples
    --------
    Create a big dense matrix ``A`` for the example:

    >>> import numpy as np
    >>> from scipy import linalg
    >>> n_rows, n_columns  = 15000, 100
    >>> rng = np.random.default_rng()
    >>> A = rng.standard_normal((n_rows, n_columns))

    Apply the transform to create a new matrix with 200 rows:

    >>> sketch_n_rows = 200
    >>> sketch = linalg.clarkson_woodruff_transform(A, sketch_n_rows, seed=rng)
    >>> sketch.shape
    (200, 100)

    Now with high probability, the true norm is close to the sketched norm
    in absolute value.

    >>> linalg.norm(A)
    1224.2812927123198
    >>> linalg.norm(sketch)
    1226.518328407333

    Similarly, applying our sketch preserves the solution to a linear
    regression of :math:`\min \|Ax - b\|`.

    >>> b = rng.standard_normal(n_rows)
    >>> x = linalg.lstsq(A, b)[0]
    >>> Ab = np.hstack((A, b.reshape(-1, 1)))
    >>> SAb = linalg.clarkson_woodruff_transform(Ab, sketch_n_rows, seed=rng)
    >>> SA, Sb = SAb[:, :-1], SAb[:, -1]
    >>> x_sketched = linalg.lstsq(SA, Sb)[0]

    As with the matrix norm example, ``linalg.norm(A @ x - b)`` is close
    to ``linalg.norm(A @ x_sketched - b)`` with high probability.

    >>> linalg.norm(A @ x - b)
    122.83242365433877
    >>> linalg.norm(A @ x_sketched - b)
    166.58473879945151

    """
    # lazy import to prevent to prevent sparse dependency for whole module (gh-23420)
    from scipy.sparse import issparse
    if issparse(input_matrix) and input_matrix.ndim > 2:
        message = "Batch support for sparse arrays is not available."
        raise NotImplementedError(message)

    S = cwt_matrix(sketch_size, input_matrix.shape[-2], rng=rng)
    if input_matrix.ndim <= 2:
        # transposes are cheap and ensure output class matches input_matrix
        return (input_matrix.T @ S.T).T

    # Despite argument order (required by decorator), this is  S @ input_matrix
    # Can avoid _batch_dot when gh-22153 is resolved.
    return _batch_dot(input_matrix, S)


@_apply_over_batch(('input_matrix', 2))
def _batch_dot(input_matrix, S):
    return S @ input_matrix
